from sarracenia.flow import Flow
import logging

logger = logging.getLogger(__name__)

default_options = {
    'acceptUnmatched': True,
    'blocksize': 1,
    'bufsize': 1024 * 1024,
    'follow_symlinks': False,
    'force_polling': False,
    'inflight': None,
    'part_ext': 'Part',
    'partflg': '1',
    'post_baseDir': None,
    'permCopy': True,
    'timeCopy': True,
    'randomize': False,
    'sumflg': 'sha512',
    'post_on_start': False,
    'sleep': 5,
    'nodupe_ttl': 0
}


class Watch(Flow):
    """
      * create messages for files that appear in a directory.

    """
    def __init__(self, options):

        super().__init__(options)
        logger.info('watching!')
        self.plugins['load'].insert(0, 'sarracenia.flowcb.gather.file.File')
        self.plugins['load'].insert(0,
                                    'sarracenia.flowcb.post.message.Message')



"""
Plugin sundewpxroute.py:
    Implement message filtering based on a routing table from MetPX-Sundew.
    Make it easier to feed clients exactly the same products with sarracenia,
    that they are used to with sundew.

Usage:
    the pxrouting option must be set in the configuration before the plugin
    is configured, like so:
    * pxRouting /local/home/peter/src/pdspx/routing/etc/pxRouting.conf
    * pxClient  navcan-amis
    flowcb sarracenia.flowcb.accept.sundewpxroute.SundewPxRoute
"""

import logging
import os

from sarracenia.flowcb import FlowCB
logger = logging.getLogger(__name__)


class SundewPxRoute(FlowCB):
    def __init__(self, options):
        """

          For World Meteorological Organization message oriented routing.
          Read the configured metpx-sundew routing table, and figure out which
          Abbreviated Header Lines (AHL's) are configured to be sent to 'target'
          being careful to account for membership in clientAliases.

          init sets 'ahls_to_route' according to the contents of pxrouting

        """
        super().__init__(options, logger)

        self.o.add_option('pxRouting', 'str', '')
        self.o.add_option('pxClient', 'str', '')

        if not self.o.pxRouting:
            logger.error("sundew_pxroute pxRouting file not defined")
            return
        elif not os.path.exists(self.o.pxRouting):
            logger.error("sundew_pxroute pxRouting file (%s) not found" % self.o.pxRouting)
            return
        
        self.ahls_to_route = {}
        
        pxrf = open(self.o.pxRouting, 'r')
        possible_references = self.o.pxClient.split(',')
        logger.info("sundew_pxroute, target clients: %s" % possible_references)

        for line in pxrf:
            words = line.split()

            if (len(words) < 2) or words[0] == '#':
                continue

            if words[0] == 'clientAlias':
                expansion = words[2].split(',')
                for i in possible_references:
                    if i in expansion:
                        possible_references.append(words[1])
                        logger.debug("sundew_pxroute adding clientAlias %s to possible_reference %s" % (words[1], possible_references))
                        continue

            if words[0] == 'key':
                expansion = words[2].split(',')
                for i in possible_references:
                    if i in expansion:
                        self.ahls_to_route[words[1]] = True
        pxrf.close()

        logger.debug("sundew_pxroute For %s, the following headers are routed %s" % (self.o.pxClient, self.ahls_to_route.keys()))

    def after_accept(self, worklist):
        new_incoming = []

        for message in worklist.incoming:
            ahl = message['new_file'].split('/')[-1][0:11]

            if (len(ahl) < 11) or (ahl[6] != '_'):
                logger.debug("sundew_pxroute not an AHL: %s" % ahl)
                worklist.rejected.append(message)
                continue

            if (ahl in self.ahls_to_route.keys()):
                logger.debug("sundew_pxroute yes, deliver: %s" % ahl)
                new_incoming.append(message)
                continue
            else:
                logger.debug("sundew_pxroute no, do not deliver: %s" % ahl)
                worklist.rejected.append(message)

        worklist.incoming = new_incoming



# This file is part of Sarracenia.
# The sarracenia suite is Free and is proudly provided by the Government of Canada
# Copyright (C) Her Majesty The Queen in Right of Canada, Shared Services Canada, 2019
#
"""
 Second version configuration parser


FIXME: pas 2023/02/05...  missing options from v2:

   max_queue_size
   outlet
   pipe

"""

import argparse
import copy
import datetime
import humanfriendly
import inspect
import logging

import os
import pathlib
import pprint
import re
import shutil
import socket
import sys
import time
import urllib, urllib.parse

from random import randint

if sys.version_info[0] >= 3 and sys.version_info[1] < 8:
    """
        'extend' action not included in argparse prior to python 3.8
        https://stackoverflow.com/questions/41152799/argparse-flatten-the-result-of-action-append
    """
 
    class ExtendAction(argparse.Action):

        def __call__(self, parser, namespace, values, option_string=None):
            items = getattr(namespace, self.dest) or []
            items.extend(values)
            setattr(namespace, self.dest, items)




import sarracenia
from sarracenia import durationToSeconds, site_config_dir, user_config_dir, user_cache_dir
from sarracenia.featuredetection import features
import sarracenia.credentials
import sarracenia.flow
import sarracenia.flowcb

from sarracenia.flow.sarra import default_options as sarradefopts

import sarracenia.identity.arbitrary

import sarracenia.moth
import sarracenia.identity
import sarracenia.instance


class octal_number(int):

    def __new__(cls, value):
        if type(value) is str:
            self = int(value,base=8)
        elif type(value) is int:
            self = value
        return self

    def __str__(self) -> str:
        return f"0o{self:o}"

    def __repr__(self) -> str:
        return f"0o{self:o}"


default_options = {
    'acceptSizeWrong': False,
    'acceptUnmatched': True,
    'amqp_consumer': False,
    'attempts': 3,
    'batch' : 100,
    'baseDir': None,
    'baseUrl_relPath': False,
    'delete': False,
    'documentRoot': None,
    'download': False,
    'dry_run': False,
    'filename': None,
    'flowMain': None,
    'inflight': None,
    'inline': False,
    'inlineOnly': False,
    'identity_method': 'sha512',
    'logMetrics': False,
    'logStdout': False,
    'nodupe_driver': 'disk',
    'nodupe_ttl': 0,
    'overwrite': True,
    'path': [],
    'permDefault' : octal_number(0),
    'permDirDefault' : octal_number(0o775),
    'permLog': octal_number(0o600),
    'post_documentRoot': None,
    'post_baseDir': None,
    'post_baseUrl': None,
    'post_format': 'v03',
    'realpathPost': False,
    'recursive' : True,
    'report': False,
    'retryEmptyBeforeExit': False,
    'sanity_log_dead': 9999,
    'sourceFromExchange': False,
    'v2compatRenameDoublePost': False,
    'varTimeOffset': 0
}

count_options = [
    'batch', 'count', 'exchangeSplit', 'instances', 'logRotateCount', 'no', 
    'post_exchangeSplit', 'prefetch', 'messageCountMax', 'messageRateMax', 
    'messageRateMin'
]


# all the boolean settings.
flag_options = [ 'acceptSizeWrong', 'acceptUnmatched', 'amqp_consumer', 'baseUrl_relPath', 'debug', \
    'delete', 'discard', 'download', 'dry_run', 'durable', 'exchangeDeclare', 'exchangeSplit', 'logReject', 'realpathFilter', \
    'follow_symlinks', 'force_polling', 'inline', 'inlineOnly', 'inplace', 'logMetrics', 'logStdout', 'logReject', 'restore', \
    'messageDebugDump', 'mirror', 'timeCopy', 'notify_only', 'overwrite', 'post_on_start', \
    'permCopy', 'queueBind', 'queueDeclare', 'randomize', 'recursive', 'realpathPost', 'reconnect', \
    'report', 'reset', 'retryEmptyBeforeExit', 'save', 'sourceFromExchange', \
    'statehost', 'users', 'v2compatRenameDoublePost'
                ]

float_options = [ ]

duration_options = [
    'expire', 'housekeeping', 'logRotateInterval', 'message_ttl', 'nodupe_fileAgeMax', 'retry_ttl',
    'sanity_log_dead', 'sleep', 'timeout', 'varTimeOffset'
]

list_options = ['path' ]

# set, valid values of the set.
set_options = [ 'logEvents', 'fileEvents' ]

set_choices = { 
    'logEvents' : sarracenia.flowcb.entry_points + [ 'reject' ],
    'fileEvents' : set( [ 'create', 'delete', 'link', 'mkdir', 'modify', 'rmdir' ] )
 }
# FIXME: doesn't work... wonder why?
#    'fileEvents': sarracenia.flow.allFileEvents
 
perm_options = [ 'permDefault', 'permDirDefault','permLog']

size_options = ['accelThreshold', 'blocksize', 'bufsize', 'byteRateMax', 'inlineByteMax']

str_options = [
    'action', 'admin', 'baseDir', 'broker', 'cluster', 'directory', 'exchange',
    'exchange_suffix', 'feeder', 'filename', 'flatten', 'flowMain', 'header', 'identity', 'logLevel', 
    'pollUrl', 'post_baseUrl', 'post_baseDir', 'post_broker', 'post_exchange',
    'post_exchangeSuffix', 'post_format', 'post_topic', 'queueName', 'sendTo', 'rename',
    'report_exchange', 'source', 'strip', 'timezone', 'nodupe_ttl', 'nodupe_driver', 
    'nodupe_basis', 'tlsRigour', 'topic', 'vip'
]
"""
   for backward compatibility, 

   convert some old plugins that are hard to get working with
   v2wrapper, into v3 plugin.

   the fdelay ones makes in depth use of sr_replay function, and
   that has changed in v3 too much.

   accelerators and rate limiting are now built-in, no plugin required.
"""
convert_to_v3 = {
    'blocksize' : ['continue'],
    'cache_stat' : ['continue'],
    'cluster_aliases' : [ 'continue' ],
    'discard' : [ 'delete_destination', 'on' ], 
    'from_cluster' : [ 'continue' ],
    'to_clusters' : [ 'continue' ],
    'identity' : {
       'n' : [ 'identity', 'none' ],
       's' : [ 'identity', 'sha512' ],
       'd' : [ 'identity', 'md5' ],
       'a' : [ 'identity', 'arbitrary' ],
       'r' : [ 'identity', 'random' ],
       'z,d' : [ 'identity', 'cod,md5' ],
       'z,s' : [ 'identity', 'cod,sha512' ],
       'z,n' : [ 'identity', 'none' ]
    },
    'ls_file_index' : [ 'continue' ],
    'plugin': {
        'msg_fdelay': ['callback', 'filter.fdelay'],
        'msg_pclean_f90':
        ['callback', 'filter.pclean_f90.PClean_F90'],
        'msg_pclean_f92':
        ['callback', 'filter.pclean_f92.PClean_F92'],
        'accel_wget': ['continue'],
        'accel_scp': ['continue'],
        'accel_cp': ['continue'],
        'msg_total_save': ['continue'],
        'post_total_save': ['continue'],
        'post_total_interval': ['continue']
    },
    'destfn_script': { 'manual_conversion_required' : [ 'continue' ] },
    'do_get': { 'manual_conversion_required' : [ 'continue' ] },
    'do_poll': { 'manual_conversion_required' : [ 'continue' ] },
    'do_put': { 'manual_conversion_required' : [ 'continue' ] },
    'do_download': { 'manual_conversion_required' : [ 'continue' ] },
    'do_put': { 'manual_conversion_required' : [ 'continue' ] },
    'do_send': {
       'file_email' : [ 'callback', 'send.email' ],
    },
    'do_task': { 'manual_conversion_required' : [ 'continue' ] },
    'no_download': [ 'download', 'False' ],
    'notify_only': [ 'download', 'False' ],
    'do_data': { 'manual_conversion_required' : [ 'continue' ] },
    'on_file': {
        'file_age' : [ 'callback', 'work.age' ],
    },
    'on_heartbeat': { 'manual_conversion_required' : [ 'continue' ] },
    'on_html_page': { 'manual_conversion_required' : [ 'continue' ] },
    'on_part': { 'manual_conversion_required' : [ 'continue' ] },
    'on_line': { 'manual_conversion_required' : [ 'continue' ] },
    'on_message': {
        'msg_print_lag': [ 'callback', 'accept.printlag.PrintLag'],
        'msg_replace_new_dir': [ 'callback', 'accept.pathreplace' ],
        'msg_skip_old': [ 'callback', 'accept.skipold.SkipOld'],
        'msg_test_retry': [ 'callback', 'accept.testretry.TestRetry'],
        'msg_to_clusters': [ 'callback', 'accept.toclusters.ToClusters'],
        'msg_save': [ 'callback', 'accept.save'],
        'msg_2localfile': [ 'callback', 'accept.tolocalfile.ToLocalFile'],
        'msg_rename_whatfn': [ 'callback', 'accept.renamewhatfn.RenameWhatFn'],
        'msg_rename_dmf': [ 'callback', 'accept.renamedmf.RenameDMF'],
        'msg_hour_tree': [ 'callback', 'accept.hourtree.HourTree'],
        'msg_renamer': [ 'callback', 'accept.renamer.Renamer'],
        'msg_2http': [ 'callback', 'accept.tohttp.ToHttp'],
        'msg_2local': [ 'callback', 'accept.tolocal.ToLocal'],
        'msg_http_to_https': [ 'callback', 'accept.httptohttps.HttpToHttps'],
        'msg_speedo': [ 'callback', 'accept.speedo.Speedo'],
        'msg_WMO_type_suffix': [ 'callback', 'accept.wmotypesuffix.WmoTypeSuffix'],
        'msg_sundew_pxroute': [ 'callback', 'accept.sundewpxroute.SundewPxRoute'],
        'msg_rename4jicc': [ 'flow_callback', 'accept.rename4jicc.Rename4Jicc'],
        'msg_delay': [ 'callback', 'accept.messagedelay.MessageDelay'],
        'msg_download_baseurl': [ 'callback', 'accept.downloadbaseurl.DownloadBaseUrl'],
	'msg_from_cluster': ['continue'],
	'msg_stdfiles': ['continue'],
	'msg_fdelay': ['callback', 'filter.fdelay'],
	'msg_stopper': ['continue'],
	'msg_overwrite_sum': ['continue'],
	'msg_gts2wistopic': ['continue'],
	'msg_download': ['continue'],
	'msg_by_source': ['continue'],
	'msg_by_user': ['continue'],
	'msg_dump': ['continue'],
	'msg_total': ['continue'],
    'on_report': { 'manual_conversion_required' : [ 'continue' ] },
    'on_stop': { 'manual_conversion_required' : [ 'continue' ] },
    'on_start': { 'manual_conversion_required' : [ 'continue' ] },
    'on_watch': { 'manual_conversion_required' : [ 'continue' ] },
    'on_post': {
        'post_log': ['logEvents', '+after_work']
    },
    'parts' : [ 'continue' ],
	'post_total': ['continue'],
        'wmo2msc': [ 'callback', 'filter.wmo2msc.Wmo2Msc'],
        'msg_delete': [ 'callback', 'filter.deleteflowfiles.DeleteFlowFiles'],
        'msg_log': ['logEvents', '+after_accept'],
        'msg_rawlog': ['logEvents', '+after_accept'],
        'post_hour_tree': [ 'callback', 'accept.posthourtree.PostHourTree'],
        'post_long_flow': [ 'callback', 'accept.longflow.LongFLow'],
        'post_override': [ 'callback', 'accept.postoverride.PostOverride'],
	'post_rate_limit': ['continue'],
        'to': ['continue']
    },
    'poll_without_vip': [ 'manual_conversion_required' ], 
    'pump' : [ 'continue' ],
    'pump_flag' : [ 'continue' ],
    'reconnect': ['continue'],
    'report_daemons': ['continue'],
    'restore' : [ 'continue' ],
    'retry_mode' : ['continue'],
    'save' : [ 'continue' ],
    'set_passwords': ['continue'],
    'windows_run': [ 'continue' ],
    'xattr_disable': [ 'continue' ]
}

# question: why don't these have matching closing braces? 
# answer: there might be an offset (-1h, -5m, etc...) and covering those cases is hard with simple substitution.

convert_patterns_to_v3 = {
   '${YYYYMMDD' : '${%Y%m%d',
   '${YYYY': '${%Y',
   '${JJJ': '${%j',
   '${HH': '${%H',
   '${DD': '${%d',

}

logger = logging.getLogger(__name__)

"""
   FIXME: respect appdir stuff using an environment variable.
   for not just hard coded as a class variable appdir_stuff

"""


def isTrue(S):
    s = S.lower()
    if s == 'true' or s == 'yes' or s == 'on' or s == '1': return True
    return False


def get_package_lib_dir():
    return os.path.dirname(inspect.getfile(Config))


def get_site_config_dir():
    return sarracenia.site_config_dir(Config.appdir_stuff['appname'],
                                   Config.appdir_stuff['appauthor'])


def get_user_cache_dir(hostdir):
    """
      hostdir = None if statehost is false, 
    """
    ucd = sarracenia.user_cache_dir(Config.appdir_stuff['appname'],
                                 Config.appdir_stuff['appauthor'])
    if hostdir:
        ucd = os.path.join(ucd, hostdir)
    return ucd


def get_user_config_dir():
    return sarracenia.user_config_dir(Config.appdir_stuff['appname'],
                                   Config.appdir_stuff['appauthor'])


def get_pid_filename(hostdir, component, configuration, no):
    """
     return the file name for the pid file for the specified instance.
   """
    piddir = get_user_cache_dir(hostdir)
    piddir += os.sep + component + os.sep

    if configuration[-5:] == '.conf':
        configuration = configuration[:-5]

    piddir += configuration + os.sep

    return piddir + os.sep + component + '_' + configuration + '_%02d' % no + '.pid'


def get_log_filename(hostdir, component, configuration, no):
    """
      return the name of a single logfile for a single instance.
   """
    logdir = get_user_cache_dir(hostdir) + os.sep + 'log'

    if configuration is None:
        configuration = ''
    else:
        configuration = '_' + configuration

    if configuration[-5:] == '.conf':
        configuration = configuration[:-5]

    return logdir + os.sep + component + configuration + '_%02d' % no + '.log'

def get_metrics_filename(hostdir, component, configuration, no):
    """
      return the name of a single logfile for a single instance.
   """
    metricsdir = get_user_cache_dir(hostdir) + os.sep + 'metrics'

    if configuration is None:
        configuration = ''
    else:
        configuration = '_' + configuration

    if configuration[-5:] == '.conf':
        configuration = configuration[:-5]

    return metricsdir + os.sep + component + configuration + '_%02d' % no + '.json'

def wget_config(urlstr, path, remote_config_url=False):
    logger.debug("wget_config %s %s" % (urlstr, path))

    try:
        req = urllib.request.Request(urlstr)
        resp = urllib.request.urlopen(req)
        if os.path.isfile(path):
            try:
                info = resp.info()
                ts = time.strptime(info.get('Last-Modified'),
                                   "%a, %d %b %Y %H:%M:%S %Z")
                last_mod_remote = time.mktime(ts)
                last_mod_local = os.stat(path).st_mtime
                if last_mod_remote <= last_mod_local:
                    logger.info("file %s is up to date (%s)" % (path, urlstr))
                    return True
            except:
                logger.error(
                    "could not compare modification dates... downloading")
                logger.debug('Exception details: ', exc_info=True)

        fp = open(path + '.downloading', 'wb')

        # top program config only needs to keep the url
        # we set option remote_config_url with the urlstr
        # at the first line of the config...
        # includes/plugins  etc... may be left as url in the config...
        # as the urlstr is kept in the config this option would be useless
        # (and damagable for plugins)

        if remote_config_url:
            fp.write(bytes("remote_config_url %s\n" % urlstr, 'utf-8'))
        while True:
            chunk = resp.read(8192)
            if not chunk: break
            fp.write(chunk)
        fp.close()

        try:
            os.unlink(path)
        except:
            pass
        os.rename(path + '.downloading', path)

        logger.info("file %s downloaded (%s)" % (path, urlstr))

        return True

    except urllib.error.HTTPError as e:
        if os.path.isfile(path):
            logger.warning('file %s could not be processed1 (%s)' %
                           (path, urlstr))
            logger.warning('resume with the one on the server')
        else:
            logger.error('Download failed 0: %s' % urlstr)
            logger.error('Server couldn\'t fulfill the request')
            logger.error('Error code: %s, %s' % (e.code, e.reason))

    except urllib.error.URLError as e:
        if os.path.isfile(path):
            logger.warning('file %s could not be processed2 (%s)' %
                           (path, urlstr))
            logger.warning('resume with the one on the server')
        else:
            logger.error('Download failed 1: %s' % urlstr)
            logger.error('Failed to reach server. Reason: %s' % e.reason)

    except Exception as e:
        if os.path.isfile(path):
            logger.warning('file %s could not be processed3 (%s) %s' %
                           (path, urlstr, e.reason))
            logger.warning('resume with the one on the server')
        else:
            logger.error('Download failed 2: %s %s' % (urlstr, e.reason))
            logger.debug('Exception details: ', exc_info=True)

    try:
        os.unlink(path + '.downloading')
    except:
        pass

    if os.path.isfile(path):
        logger.warning("continue using existing %s" % path)

    return False


def config_path(subdir, config, mandatory=True, ctype='conf'):
    """
    Given a subdir/config look for file in configish places.

    return Tuple:   Found (True/False), path_of_file_found|config_that_was_not_found
    """
    logger.debug("config_path = %s %s" % (subdir, config))

    if config == None: return False, None

    # remote config

    if config.startswith('http:'):
        urlstr = config
        name = os.path.basename(config)
        if not name.endswith(ctype): name += '.' + ctype
        path = get_user_config_dir() + os.sep + subdir + os.sep + name
        config = name

        logger.debug("http url %s path %s name %s" % (urlstr, path, name))

        # do not allow plugin (Peter's mandatory decision)
        # because plugins may need system or python packages
        # that may not be installed on the current server.
        if subdir == 'plugins':
            logger.error("it is not allowed to download plugins")
        else:
            ok = Config.wget_config(urlstr, path)

    # priority 1 : config given is a valid path

    logger.debug("config_path %s " % config)
    if os.path.isfile(config):
        return True, config
    config_file = os.path.basename(config)
    config_name = re.sub(r'(\.inc|\.conf|\.py)', '', config_file)
    ext = config_file.replace(config_name, '')
    if ext == '': ext = '.' + ctype
    config_path = config_name + ext

    # priority 1.5: config file given without extenion...
    if os.path.isfile(config_path):
        return True, config_path

    # priority 2 : config given is a user one

    config_path = os.path.join(get_user_config_dir(), subdir,
                               config_name + ext)
    logger.debug("config_path %s " % config_path)

    if os.path.isfile(config_path):
        return True, config_path

    # priority 3 : config given to site config

    config_path = os.path.join(get_site_config_dir(), subdir,
                               config_name + ext)
    logger.debug("config_path %s " % config_path)

    if os.path.isfile(config_path):
        return True, config_path

    # priority 4 : plugins

    if subdir == 'plugins':
        config_path = get_package_lib_dir(
        ) + os.sep + 'plugins' + os.sep + config_name + ext
        logger.debug("config_path %s " % config_path)
        if os.path.isfile(config_path):
            return True, config_path

    # return bad file ...
    if mandatory:
        if subdir == 'plugins': logger.error("script not found %s" % config)
        elif config_name != 'plugins':
            logger.error("file not found %s" % config)

    return False, config


class Config:
    """
       The option parser to produce a single configuration.

       it can be instantiated with one of:

       * one_config(component, config, isPost=False) -- read the options  for
         a given component an configuration,  (all in one call.)

       On the other hand, a configu can be built up from the following constructors:

       * default_config() -- returns an empty configuration, given a config file tree.
       * no_file_config() -- returns an empty config without any config file tree. 

       Then just add settings manually::

         cfg = no_file_config()

         cfg.broker = sarracenia.credentials.Credential('amqps://anonymous:anonymous@hpfx.collab.science.gc.ca')
         cfg.topicPrefix = [ 'v02', 'post']
         cfg.component = 'subscribe'
         cfg.config = 'flow_demo'
         cfg.bindings = [ ('xpublic', ['v02', 'post'], ['*', 'WXO-DD', 'observations', 'swob-ml', '#' ]) ]
         cfg.queueName='q_anonymous.subscriber_test2'
         cfg.download=True
         cfg.batch=1
         cfg.messageCountMax=5
       
         # set the instance number for the flow class.
         cfg.no=0
       
       # and at the end call finalize

       cfg.finalize()

    """

    port_required = [ 'on_line', 'on_html_page' ]

    v2entry_points = [
        'do_download', 'do_get', 'do_poll', 'do_put', 'do_send', 'on_message',
        'on_file', 'on_heartbeat', 'on_housekeeping',
        'on_html_page', 'on_part', 'on_post', 'on_report',
        'on_start', 'on_stop', 'on_watch', 'plugin'
    ]
    components = [
        'audit', 'cpost', 'cpump', 'flow', 'poll', 'post', 'sarra', 'sender', 'shovel',
        'subscribe', 'watch', 'winnow'
    ]

    actions = [
        'add', 'cleanup', 'convert', 'devsnap', 'declare', 'disable', 'dump', 'edit',
        'enable', 'features', 'foreground', 'log', 'list', 'remove', 'restart', 'run', 'sanity',
        'setup', 'show', 'start', 'stop', 'status', 'overview'
    ]

    # lookup in dictionary, respond with canonical version.
    appdir_stuff = {'appauthor': 'MetPX', 'appname': 'sr3'}

    # Correct name on the right, old name on the left.
    synonyms = {
        'a': 'action',
        'accel_cp_threshold': 'accelThreshold',
        'accel_scp_threshold': 'accelThreshold',
        'accel_wget_threshold': 'accelThreshold',
        'accept_unmatch': 'acceptUnmatched',
        'accept_unmatched': 'acceptUnmatched',
        'at': 'attempts', 
        'b': 'broker',
        'bd': 'baseDir',
        'basedir': 'baseDir',
        'base_dir': 'baseDir',
        'baseurl': 'baseUrl',
        'bind_queue': 'queueBind',
        'cache': 'nodupe_ttl',
        'c': 'include',
        'cb': 'nodupe_basis',
        'cache_basis': 'nodupe_basis',
        'caching': 'nodupe_ttl',
        'chmod': 'permDefault',
        'chmod_dir': 'permDirDefault',
        'chmod_log': 'permLog',
        'content' : 'inline', 
        'content_encoding':  'inlineEncoding',
        'content_max': 'inlineByteMax',
        'd': 'discard',
        'declare_exchange': 'exchangeDeclare',
        'declare_queue': 'queueDeclare',
        'default_mode': 'permDefault',
        'default_dir_mode': 'permDirDefault',
        'default_log_mode': 'permLog',
        'destination_timezone': 'timezone', 
        'document_root': 'documentRoot',
        'download-and-discard': 'discard',
        'e' : 'fileEvents',
        'events' : 'fileEvents',
        'ex': 'exchange',
        'exchange_split': 'exchangeSplit',
        'exchange_suffix': 'exchangeSuffix',
        'expiry': 'expire', 
        'file_time_limit' : 'nodupe_fileAgeMax', 
        'fp' : 'force_polling',
        'fs' : 'follow_symlinks',
        'h' : 'help',
        'heartbeat': 'housekeeping',
        'hb_memory_baseline_file' : 'MemoryBaseLineFile',
        'hb_memory_max' : 'MemoryMax',
        'hb_memory_multiplier' : 'MemoryMultiplier',
        'imx': 'inlineByteMax',
        'inl' : 'inline', 
        'inline_encoding':  'inlineEncoding',
        'inline_max': 'inlineByteMax',
        'instance': 'instances',
        'lock': 'inflight', 
        'log_format': 'logFormat',
        'll': 'logLevel',
        'loglevel': 'logLevel',
        'log_reject': 'logReject',
        'logdays': 'logRotateCount',
        'log_rotate': 'logRotateCount',
        'logRotate': 'logRotateCount',
        'logRotate': 'logRotateCount',
        'logRotate_interval': 'logRotateInterval',
        'message-ttl': 'message_ttl',
        'msg_replace_new_dir' : 'pathReplace',
        'msg_filter_wmo2msc_replace_dir': 'filter_wmo2msc_replace_dir',
        'msg_filter_wmo2msc_uniquify': 'filter_wmo2msc_uniquify',
        'msg_filter_wmo2msc_tree': 'filter_wmo2msc_treeify',
        'msg_filter_wmo2msc_convert': 'filter_wmo2msc_convert',
        'msg_fdelay' : 'fdelay',
        'n': 'no_download', 
        'nd': 'nodupe_ttl',
        'no_duplicates': 'nodupe_ttl',
        'o' : 'overwrite', 
        'on_msg': 'on_message',
        'p' : 'path',
        'pm' : 'permCopy',
        'post_base_dir': 'post_baseDir',
        'post_basedir': 'post_baseDir',
        'post_base_url': 'post_baseUrl',
        'post_baseurl': 'post_baseUrl',
        'post_document_root': 'post_documentRoot',
        'post_exchange_split': 'post_exchangeSplit',
        'post_exchange_suffix': 'post_exchangeSuffix',
        'post_rate_limit': 'messageRateMax',
        'post_topic_prefix' : 'post_topicPrefix',
        'preserve_mode' : 'permCopy',
        'preserve_time' : 'timeCopy',
        'pt' : 'timeCopy',
        'qn': 'queueName',
        'queue' : 'queueName', 
        'queue_name' : 'queueName', 
        'realpath' : 'realpathPost',
        'realpath_filter' : 'realpathFilter',
        'realpath_post' : 'realpathPost',
        'remoteUrl' : 'sendTo', 
        'report_back': 'report',
        'sd' : 'nodupe_ttl',
        'sdb' : 'nodupe_basis', 
        'simulate': 'dry_run',
        'simulation': 'dry_run',
        'source_from_exchange': 'sourceFromExchange', 
        'sum' : 'identity',  
        'suppress_duplicates' : 'nodupe_ttl',
        'suppress_duplicates_basis' : 'nodupe_basis', 
        'tls_rigour' : 'tlsRigour', 
        'topic_prefix' : 'topicPrefix'
    }
    credentials = None

    def __init__(self, parent=None) -> 'Config':
        """
          instantiate an empty Configuration
        """
        self.bindings = []
        self.__admin = None
        self.__broker = None
        self.__post_broker = None

        if Config.credentials is None:
            Config.credentials = sarracenia.credentials.CredentialDB()
            Config.credentials.read(get_user_config_dir() + os.sep +
                                    "credentials.conf")
        self.directory = None

        self.env = copy.deepcopy(os.environ)

        egdir = os.path.dirname(inspect.getfile(sarracenia.config.Config)) + os.sep + 'examples' 

        self.config_search_path = [ "." , get_user_config_dir(), egdir, egdir + os.sep + 'flow'  ]


        for k in default_options:
            setattr(self, k, default_options[k])

        if parent is not None:
            for i in parent:
                setattr(self, i, parent[i])

        self.bufsize = 1024 * 1024
        self.byteRateMax = 0

        self.nodupe_fileAgeMax = 0 # disabled.
        self.timezone = 'UTC'
        self.debug = False
        self.declared_exchanges = []
        self.discard = False
        self.displayFull = False
        self.dry_run = False
        self.env_declared = []  # list of variable that are "declared env"'d 
        self.v2plugins = {}
        self.v2plugin_options = []
        self.imports = []
        self.logEvents = set(['after_accept', 'after_post', 'after_work', 'on_housekeeping' ])
        self.destfn_scripts = []
        self.plugins_late = []
        self.plugins_early = []
        self.exchange = None
        self.filename = None
        self.fixed_headers = {}
        self.flatten = '/'
        self.hostname = socket.getfqdn()
        self.hostdir = socket.getfqdn().split('.')[0]
        self.log_flowcb_needed = False
        self.sleep = 0.1
        self.housekeeping = 300
        self.inline = False
        self.inlineByteMax = 4096
        self.inlineEncoding = 'guess'
        self.identity_arbitrary_value = None
        self.logReject = False
        self.logRotateCount = 5
        self.logRotateInterval = 60*60*24
        self.masks = []
        self.instances = 1
        self.mirror = False
        self.messageAgeMax = 0
        self.post_exchanges = []
	    #self.post_topicPrefix = None
        self.pstrip = False
        self.queueName = None
        self.randomize = False
        self.rename = None
        self.randid = "%04x" % randint(0, 65536)
        self.statehost = False
        self.settings = {}
        self.strip = 0
        self.timeout = 300
        self.tlsRigour = 'normal'
        self.topicPrefix = [ 'v03', 'post' ]
        self.undeclared = []
        self.declared_users = {}
        self.users = False
        self.vip = None

    def __deepcopy__(self, memo) -> 'Configuration':
        """
            code for this from here: https://stackoverflow.com/questions/1500718/how-to-override-the-copy-deepcopy-operations-for-a-python-object
            Needed for python < 3.7ish? (ubuntu 18) found this bug: https://bugs.python.org/issue10076
            deepcopy fails for objects with re's in them?
            ok on ubuntu 20.04
        """
        cls = self.__class__
        result = cls.__new__(cls)
        memo[id(self)] = result
        for k, v in self.__dict__.items():
            if k == 'masks':
                v2=[]
                for m in v:
                    v2.append(tuple(list(copy.deepcopy(m[0:3]))+ [m[3]] + list(copy.deepcopy(m[4:]))))

                setattr(result, k, v2)
            else:
                setattr(result, k, copy.deepcopy(v, memo))
        return result

    def _validate_urlstr(self, urlstr) -> tuple :
        """
           returns a tuple ( bool, expanded_url ) 
           the bool is whether the expansion worked, and the expanded_url is one with
           the added necessary authentication details from sarracenia.Credentials.

        """
        # check url and add credentials if needed from credential file
        ok, cred_details = Config.credentials.get(urlstr)
        if cred_details is None:
            logging.critical("bad credential %s" % urlstr)
            # Callers expect that a Credential object will be returned
            cred_details = sarracenia.credentials.Credential()
            cred_details.url = urllib.parse.urlparse(urlstr)
            return False, cred_details
        return True, cred_details

    def applyComponentDefaults( self, component ):
        """
          overlay defaults options for the given component to the given configuration.
        """
        if component in ['post']:
            self.override(sarracenia.flow.post.default_options)
        elif component in ['poll']:
            self.override(sarracenia.flow.poll.default_options)
        elif component in ['sarra']:
            self.override(sarradefopts)
        elif component in ['sender']:
            self.override(sarracenia.flow.sender.default_options)
        elif component in ['subscribe']:
            self.override(sarracenia.flow.subscribe.default_options)
        elif component in ['watch']:
            self.override(sarracenia.flow.watch.default_options)

    @property
    def admin(self):
        return self.__admin

    @admin.setter
    def admin(self, v):
        if type(v) is str:
            ok, cred_details = self._validate_urlstr(v)
            if ok:
                self.__admin = cred_details
        else:
            self.__admin = v

    @property
    def broker(self):
        return self.__broker

    @broker.setter
    def broker(self, v):
        if type(v) is str:
            ok, cred_details = self._validate_urlstr(v)
            if ok:
                self.__broker = cred_details
        else:
            self.__broker = v

    @property
    def post_broker(self):
        return self.__post_broker

    @post_broker.setter
    def post_broker(self, v):
        if type(v) is str:
            ok, cred_details = self._validate_urlstr(v)
            if ok:
                self.__post_broker = cred_details
        else:
            self.__post_broker = v

    def _varsub(self, word):
        """ substitute variable values from options
       """

        if word is None:
            return word
        elif type(word) in [bool, int, float, octal_number]:
            return word
        elif not '$' in word:
            return word

        result = word
        if (('${BROKER_USER}' in word) and hasattr(self, 'broker') and self.broker is not None and
                self.broker.url is not None and hasattr(self.broker.url, 'username')):
            result = result.replace('${BROKER_USER}', self.broker.url.username)
            # FIXME: would this work also automagically if BROKER.USERNAME ?

        if (('${POST_BROKER_USER}' in word) and hasattr(self, 'post_broker') and self.post_broker is not None and
                self.post_broker.url is not None and hasattr(self.post_broker.url, 'username')):
            result = result.replace('${POST_BROKER_USER}', self.post_broker.url.username)

        if not '$' in result:
            return result

        elst = []
        plst = result.split('}')
        for parts in plst:
            try:
                if '{' in parts: elst.append((parts.split('{'))[1])
            except:
                pass
        for E in elst:
            if E in ['PROGRAM']:
                e = 'component'
            else:
                e = E.lower()
            if hasattr(self, e):
                repval = getattr(self, e)
                if type(repval) is list:
                    repval = repval[0]
                result = result.replace('${' + E + '}', repval)
                continue

            if E in self.env.keys():
                result = result.replace('${' + E + '}', self.env[E])
                if sys.platform == 'win32':
                    result = result.replace('\\', '/')
        return result

    def _build_mask(self, option, arguments):
        """ return new entry to be appended to list of masks
       """
        try:
            regex = re.compile(arguments[0])
        except:
            logger.critical( f"invalid regular expression: {arguments[0]}, ignored." )
            return None

        if len(arguments) > 1:
            fn = arguments[1]
        else:
            fn = self.filename
        if fn and re.compile('DESTFNSCRIPT=.*').match(fn):
            script=fn[13:]
            self.destfn_scripts.append(script)

        if self.directory:
           d = os.path.expanduser(self.directory)
        else:
           d = self.directory
        return (arguments[0], d, fn, regex,
                option.lower() in ['accept' ], self.mirror, self.strip,
                self.pstrip, self.flatten)

    def mask_ppstr(self, mask):
        """
           return a pretty print string version of the given mask, easier for humans to read.
        """
        pattern, maskDir, maskFileOption, mask_regexp, accepting, mirror, strip, pstrip, flatten = mask

        s = 'accept' if accepting else 'reject'
        if pstrip : strip=pstrip
        strip = '' if strip == 0 else f' strip:{strip}'
        fn = '' if (maskFileOption == 'WHATFN') else f' filename:{maskFileOption}'
        flatten = '' if flatten == '/' else f' flatten:{flatten}'
        w = 'with ' if fn or flatten or strip else ''
        return f'{s} {pattern} into {maskDir} {w}mirror:{mirror}{strip}{flatten}{fn}'

    def add_option(self, option, kind='list', default_value=None, all_values=None ):
        """
           options can be declared in any plugin. There are various *kind* of options, where the declared type modifies the parsing.
           
           * 'count'      integer count type. 

           * 'octal'      base-8 (octal) integer type.
           * 'duration'   a floating point number indicating a quantity of seconds (0.001 is 1 milisecond)
                          modified by a unit suffix ( m-minute, h-hour, w-week ) 

           * 'flag'       boolean (True/False) option.

           * 'float'      a simple floating point number.

           * 'list'       a list of string values, each succeeding occurrence catenates to the total.
                          all v2 plugin options are declared of type list.

           * 'set'        a set of string values, each succeeding occurrence is unioned to the total.
                          if all_values is provided, then constrain set to that.

           * 'size'       integer size. Suffixes k, m, and g for kilo, mega, and giga (base 2) multipliers.

           * 'str'        an arbitrary string value, as will all of the above types, each 
                          succeeding occurrence overrides the previous one.
    
           If a value is set to None, that could mean that it has not been set.
        """
        #Blindly add the option to the list if it doesn't already exist
        if not hasattr(self, option):
            setattr(self, option, default_value)

        # Retreive the 'new' option & enforce the correct type.
        v = getattr(self, option)

        if kind not in [ 'list', 'set' ] and type(v) == list:
            v=v[-1]
            logger.warning( f"multiple declarations of {option}={getattr(self,option)} choosing last one: {v}" )


        if kind == 'count':
            count_options.append(option)
            if type(v) is not int:
                setattr(self, option, humanfriendly.parse_size(v))
        elif kind == 'duration':
            duration_options.append(option)
            if type(v) is not float:
                setattr(self, option, durationToSeconds(v,default_value))
        elif kind == 'flag':
            flag_options.append(option)
            if type(v) is not bool:
                setattr(self, option, isTrue(v))
        elif kind == 'float':
            float_options.append(option)
            if type(v) is not float:
                setattr(self, option, float(v))
        elif kind == 'list':  
            list_options.append( option )
            if type(v) is not list:
                #subtlety... None means: has not been set, 
                # where an empty list to be an explicit setting.
                if v is None:
                    setattr(self, option, None)
                else:
                    setattr(self, option, [v])
        elif kind == 'octal':
            perm_options.append(option)
            if type(v) is not octal_number:
                setattr(self, option, octal_number(int(v,base=8)))
        elif kind == 'set':  
            set_options.append(option)
            sv=set()
            if type(v) is list:
                sv=set(v)
            elif type(v) is set:
                sv=v
            elif type(v) is str:
                if v == 'None': 
                    sv=set([])
                else:
                    if v[0] in [ '+', '-']:
                        op=v[0]
                        v=v[1:]
                    else:
                        op='r'

                    if ',' in v: 
                        sv=set(v[1:].split(','))
                    else: 
                        sv=set([v[1:]])

                    if op == '+':
                        sv= getattr(self,option) | sv
                    elif op == '-' :
                        sv= getattr(self,option) - sv
                    else:
                        sv=set([v])
            setattr(self, option, sv)
            if all_values:
                set_choices[option] = all_values

        elif kind == 'size':
            size_options.append(option)
            if type(v) is not int:
                setattr(self, option, humanfriendly.parse_size(v))

        elif kind == 'str':
            str_options.append(option)
            if v is None:
                setattr(self, option, None)
            elif type(v) is not str:
                setattr(self, option, str(v))
        else:
            logger.error('invalid kind: %s for option: %s, ignored' % ( kind, option ) )
            return

        logger.debug('%s declared as type:%s value:%s' % (option, type(getattr(self,option)), v))

    def dump(self):
        """ print out what the configuration looks like.
       """

        term = shutil.get_terminal_size((80, 20))

        # for python > 3.7
        #c = copy.deepcopy(self.dictify())
        # but older python needs:
        c = self.dictify()
        d={}
        for k in c:
            if k == 'masks':
                i=0
                d['masks'] = []
                while i < len(c['masks']):
                   d['masks'].append( self.mask_ppstr(c['masks'][i]) )
                   i+=1
            else:
                d[k] = copy.deepcopy(c[k])

        for omit in [ 'env' ] :
            del d[omit]

        for k in d:
            if type(d[k]) is sarracenia.credentials.Credential :
                d[k] = str(d[k])

        pprint.pprint( d, width=term.columns, compact=True )
        return

    def dictify(self):
        """
      return a dict version of the cfg... 
      """
        cd = self.__dict__

        if hasattr(self, 'admin'):
            cd['admin'] = self.admin

        if hasattr(self, 'broker'):
            cd['broker'] = self.broker

        if hasattr(self, 'post_broker'):
            cd['post_broker'] = self.post_broker

        return cd

    
    def get_source_from_exchange(self,exchange):
        #self.logger.debug("%s get_source_from_exchange %s" % (self.program_name,exchange))

        source = None
        if len(exchange) < 4 or not exchange.startswith('xs_') : return source

        # check if source is a valid declared source user

        len_u   = 0
        try:
                # look for user with role source
                for u in self.declared_users :
                    if self.declared_users[u] != 'source' : continue
                    if exchange[3:].startswith(u) and len(u) > len_u :
                       source = u
                       len_u  = len(u)
        except: pass

        return source

 

    def _merge_field(self, key, value):
        if key == 'masks':
            self.masks += value
        else:
            if value is not None:
                setattr(self, key, value)

    def merge(self, oth):
        """ 
       merge to lists of options.

       merge two lists of options if one is cumulative then merge, 
       otherwise if not None, then take value from oth
       """

        if type(oth) == dict:
            for k in oth.keys():
                self._merge_field(k, self._varsub(oth[k]))
        else:
            for k in oth.__dict__.keys():
                self._merge_field(k, self._varsub(getattr(oth, k)))

    def _override_field(self, key, value):
        if key == 'masks':
            self.masks += value
        else:
            setattr(self, key, value)

    def override(self, oth):
        """
       override a value in a set of options.

       why override() method and not just assign values to the dictionary?
       in the configuration file, there are various ways to have variable substituion.
       override invokes those, so that they are properly interpreted.  Otherwise,
       you just end up with a literal value.
       """

        if type(oth) == dict:
            for k in oth.keys():
                self._override_field(k, self._varsub(oth[k]))
        else:
            for k in oth.__dict__.keys():
                logger.error('k=%s, v=%s' % (k, getattr(oth, k)))
                self._override_field(k, self._varsub(getattr(oth, k)))

    def _resolve_exchange(self):
        """
           based on the given configuration, fill in with defaults or guesses.
           sets self.exchange.
        """
        if not hasattr(self, 'exchange') or self.exchange is None:
            #if hasattr(self, 'post_broker') and self.post_broker is not None and self.post_broker.url is not None:
            #    self.exchange = 'xs_%s' % self.post_broker.url.username
            #else:
            if not hasattr(self.broker.url,'username') or ( self.broker.url.username == 'anonymous' ):
                self.exchange = 'xpublic'
            else:
                self.exchange = 'xs_%s' % self.broker.url.username

            if hasattr(self, 'exchangeSuffix'):
                self.exchange += '_%s' % self.exchangeSuffix

            if hasattr(self, 'exchangeSplit') and hasattr(
                    self, 'no') and (self.no > 0):
                self.exchange += "%02d" % self.no

    def _parse_binding(self, subtopic_string):
        """
         FIXME: see original parse, with substitions for url encoding.
                also should sqwawk about error if no exchange or topicPrefix defined.
                also None to reset to empty, not done.
       """
        if hasattr(self, 'broker') and self.broker is not None and self.broker.url is not None:
            self._resolve_exchange()

        if type(subtopic_string) is str:
            if not hasattr(self, 'broker') or self.broker is None or self.broker.url is None:
                logger.error( 'broker needed before subtopic' )
                return

            if self.broker.url.scheme == 'amq' :
                subtopic = subtopic_string.split('.')
            else:
                subtopic = subtopic_string.split('/')
            
        if hasattr(self, 'exchange') and hasattr(self, 'topicPrefix'):
            self.bindings.append((self.exchange, self.topicPrefix, subtopic))

    def _parse_v2plugin(self, entryPoint, value):
        """
       config file parsing for a v2 plugin.

       """
        if not entryPoint in Config.v2entry_points:
            logging.error(
                "undefined entry point: {} skipped".format(entryPoint))
            return

        if not entryPoint in self.v2plugins:
            self.v2plugins[entryPoint] = [value]
        else:
            self.v2plugins[entryPoint].append(value)

    def _parse_declare(self, words):

        if words[0] in ['env', 'envvar', 'var', 'value']:
            name, value = words[1].split('=')
            self.env[name] = value
            self.env_declared.append(name)
        elif words[0] in ['option', 'o']:
            self._parse_option(words[1], words[2:])
        elif words[0] in ['source', 'subscriber', 'subscribe']:
            self.declared_users[words[1]] = words[0]
        elif words[0] in ['exchange']:
            self.declared_exchanges.append(words[1])

    def _parse_setting(self, opt, value):
        """
          v3 plugin accept options for specific modules.
    
          parsed from:
          set sarracenia.flowcb.log.filter.Log.level debug

          example:   
          opt= sarracenia.flowcb.log.filter.Log.level  value = debug

          results in:
          self.settings[ sarracenia.flowcb.log.filter.Log ][level] = debug

          options should be fed to plugin class on instantiation.
          stripped of class... 
               options = { 'level' : 'debug' }
    

       """
        opt_class = '.'.join(opt.split('.')[:-1])
        opt_var = opt.split('.')[-1]
        if opt_class not in self.settings:
            self.settings[opt_class] = {}

        self.settings[opt_class][opt_var] = ' '.join(value)

    def _parse_sum(self, value):
        #logger.error('FIXME! input value: %s' % value)

        if not value:
            if not self.identity_method:
               return
            value = self.identity_method

        if (value in sarracenia.identity.known_methods) or (
                value[0:4] == 'cod,'):
            self.identity_method = value
            #logger.error('returning 1: %s' % value)
            return

        #logger.error( f'1 value: {value} self.identity_method={self.identity_method}' )
        if (value[0:2] == 'z,'):
            value = value[2:]
            self.identity_method = 'cod,'
        elif (value[0:2] == 'a,'):
            self.identity_method = 'arbitrary' 
            self.identity_arbitrary_value = value[2:]
        else:
            self.identity_method = value
        #logger.error( f'2 value: {value} self.identity_method={self.identity_method}' )

        if value.lower() in [ 'n', 'none' ]:
            self.identity_method = None
            #logger.error('returning 1.1: %s' % 'none')
            return 
        #logger.error( f'3 value: {value} self.identity_method={self.identity_method}' )

        for sc in sarracenia.identity.Identity.__subclasses__():
            #logger.error('against 1.8: %s' % sc.__name__.lower() )
            if value == sc.__name__.lower():
                #logger.error('returning 2: %s' % value )
                if self.identity_method == 'cod,':
                      self.identity_method += value
                else:
                      self.identity_method = value
                return
            if hasattr(sc, 'registered_as'):
                #logger.error('against 3: %s' % sc.registered_as() )

                if (sc.registered_as() == value):
                    if self.identity_method == 'cod,':
                          self.identity_method += sc.__name__.lower()
                    else:
                          self.identity_method = sc.__name__.lower()
                    #logger.error('returning 3: %s' % self.identity_method)
                    return
        # FIXME this is an error return case, how to designate an invalid checksum?
        self.identity_method = 'invalid'
        #logger.error('returning 4: invalid' )

    def parse_file(self, cfg, component=None):
        """ add settings from a given config file to self 
       """
        if component:
            cfname = f'{component}/{cfg}'
        else:
            cfname = cfg

        logger.debug( f'looking for {cfg} (in {os.getcwd()}')

        cfg=os.path.expanduser(cfg)

        if cfg[0] == os.sep:
            cfgfilepath=cfg
        else:
            cfgfilepath=None
            for d in self.config_search_path:
                 cfgfilepath=d + os.sep + cfg
                 if os.path.isfile( cfgfilepath ):
                     break

            if not cfgfilepath:
                 logger.error( f'failed to find {cfg}' )
                 return
            logger.debug( f'found {cfgfilepath}')

        lineno=0
        for l in open(cfgfilepath, "r").readlines():
            l = l.strip()
            lineno+=1
            line = l.split()

            #print('FIXME parsing %s:%d %s' % (cfg, lineno, line ))

            if (len(line) < 1) or (line[0].startswith('#')):
                continue

            k = line[0]
            if k in Config.synonyms:
                k = Config.synonyms[k]
            elif k == 'destination':
                if component == 'poll':
                    k = 'pollUrl'
                else:
                    k = 'sendTo'
            elif k == 'broker' and component == 'poll' :
                k = 'post_broker'
            elif k == 'directory' and component == 'poll' :
                k = 'path'

            if (k in convert_to_v3): 
                self.log_flowcb_needed |= '_log' in k
                   
                if (len(line) > 1):
                    v = line[1].replace('.py', '', 1)
                    if (v in convert_to_v3[k]):
                        line = convert_to_v3[k][v]
                        k = line[0]
                        if 'continue' in line:
                            logger.debug( f'{cfname}:{lineno} obsolete v2: \"{l}\" ignored' )
                        else:
                            logger.debug( f'{cfname}:{lineno} obsolete v2:\"{l}\" converted to sr3:\"{" ".join(line)}\"' )
                else:
                    line = convert_to_v3[k]
                    k=line[0]
                    v=line[1] 

            if k == 'continue':
                continue
            
            #FIXME: note for Clea, line conversion to v3 complete here.

            line = list(map(lambda x: self._varsub(x), line))

            if len(line) == 1:
                v = True
            else:
                v = line[1]

            # FIXME... I think synonym check should happen here, but no time to check right now.

            if k in flag_options:
                if len(line) == 1:
                    setattr(self, k, True)
                else:
                    setattr(self, k, isTrue(v))
            
                if k in ['logReject'] and self.logReject:
                    self.logEvents = self.logEvents | set(['reject'])
                continue

            if len(line) < 2:
                logger.error('%s:%d %s missing argument(s) ' % ( cfname, lineno, k ) )
                continue
            if k in ['accept', 'reject' ]:
                self.masks.append(self._build_mask(k, line[1:]))
            elif k in [ 'callback', 'cb' ]:
                #vv = v.split('.')
                #v = 'sarracenia.flowcb.' + v + '.' + vv[-1].capitalize()
                if v not in self.plugins_late:
                    self.plugins_late.append(v)
            elif k in [ 'callback_prepend', 'cbp' ]:
                #vv = v.split('.')
                #v = 'sarracenia.flowcb.' + v + '.' + vv[-1].capitalize()
                if v not in self.plugins_early:
                    self.plugins_early.insert(0,v)
            elif k in ['declare']:
                self._parse_declare(line[1:])
            elif k in ['feeder', 'manager']:
                self.feeder = urllib.parse.urlparse(line[1])
                self.declared_users[self.feeder.username] = 'feeder'
            elif k in ['header', 'h']:
                (kk, vv) = line[1].split('=')
                self.fixed_headers[kk] = vv
            elif k in ['include', 'config']:
                try:
                    self.parse_file(v)
                except Exception as ex:
                    logger.error('file %s failed to parse:  %s' % (v, ex))
                    logger.debug('Exception details: ', exc_info=True)
            elif k in ['subtopic']:
                self._parse_binding(v)
            elif k in ['topicPrefix']:
                if '/' in v :
                    self.topicPrefix = v.split('/')
                else:
                    self.topicPrefix = v.split('.')
            elif k in ['post_topicPrefix']:
                #if (not self.post_broker.url) or self.post_broker.url.scheme[0:3] == 'amq':
                if '/' in v :
                    self.post_topicPrefix = v.split('/')
                else:
                    self.post_topicPrefix = v.split('.')
            elif k in ['import']:
                self.imports.append(v)
            elif k in ['flow_callback', 'flowcb', 'fcb', 'flowCallback' ]:
                if v not in self.plugins_late:
                    self.plugins_late.append(v)
            elif k in ['flow_callback_prepend', 'flowcb_prepend', 'fcbp', 'flowCallbackPrepend' ]:
                if v not in self.plugins_early:
                    self.plugins_early.insert(0, v)
            elif k in ['set', 'setting', 's']:
                self._parse_setting(line[1], line[2:])
            elif k in ['identity', 'integrity']:
                self._parse_sum(v)
            elif k in Config.port_required:
                logger.error( f' {cfname}:{lineno} {k} {v} not supported in v3, consult porting guide. Option ignored.' )
                logger.error( f' porting guide: https://github.com/MetPX/sarracenia/blob/v03_wip/docs/How2Guides/v2ToSr3.rst ' )
                continue
            elif k in Config.v2entry_points:
                #if k in self.plugins:
                #    self.plugins.remove(v)
                self._parse_v2plugin(k, v)
            elif k in ['no-import']:
                self._parse_v3unplugin(v)
            elif k in ['inflight', 'lock']:
                if v[:-1].isnumeric():
                    setattr(self, k, durationToSeconds(v))
                else:
                    if line[1].lower() in ['none', 'off', 'false']:
                        setattr(self, k, None)
                    else:
                        setattr(self, k, v)
            elif k in ['strip']:
                """
               2020/08/26 - PAS
               strip in config file gets translated into two separate attributes: strip and pstrip.
                 strip is the numeric variety (0-n) and if the supplied option in a regex pattern, 
                 then instead pstrip is set, and strip is set to 0.

               I don't know why it is done this way... just documenting/conforming to existing state.
               """
                if v.isdigit():
                    self.strip = int(v)
                    self.pstrip = None
                else:
                    if v[0] == '/':
                        self.pstrip = v[1:]
                    else:
                        self.pstrip = v
                    self.strip = 0
            elif k in duration_options:
                if len(line) == 1:
                    logger.error( 
                        '%s:%d  %s is a duration option requiring a decimal number of seconds value'
                        % ( cfname, lineno, line[0]) )
                    continue
                setattr(self, k, durationToSeconds(v))
            elif k in float_options:
                try:
                    setattr(self, k, float(v))
                except (ValueError, TypeError) as e:
                    logger.error(f'Ignored "{i}": {e}')
            elif k in perm_options:
                if v.isdigit():
                    setattr(self, k, octal_number(int(v, base=8)))
                else:
                    logger.error('%s setting to %s ignored: only numberic modes supported' % ( k, v ) )
            elif k in size_options:
                setattr(self, k, humanfriendly.parse_size(v))
            elif k in count_options:
                setattr(self, k, humanfriendly.parse_size(v))
            elif k in list_options:
                if not hasattr(self, k):
                    setattr(self, k, [' '.join(line[1:])])
                else:
                    getattr(self, k).append(' '.join(line[1:]))
            elif k in set_options:
                if v.lower() == 'none':
                    setattr(self, k, set([]))
                    continue
                if v.lower() in [ 'all' , '+all' ]:
                    if k in set_choices:
                        setattr(self,k,set_choices[k])
                    continue
                v=v.replace('|',',')
                if v[0] in ['-','+']:
                    op=v[0]
                    v=v[1:]
                else:
                    op='r'

                if op == '+':
                    vs = getattr(self,k) | set(v.split(','))
                elif op == '-':
                    vs = getattr(self,k) - set(v.split(','))
                else:
                    vs = set( v.split(',') )

                setattr(self, k, vs )

                if k in set_choices :
                    for i in getattr(self,k):
                        if i not in set_choices[k]:
                            logger.error('invalid entry for %s:  %s. Must be one of: %s' % ( k, i, set_choices[k] ) )

            elif k in str_options:
                v = ' '.join(line[1:])
                if v == 'None':
                    v=None
                setattr(self, k, v)
            else:
                #FIXME: with _options lists for all types and addition of declare, this is probably now dead code.
                if k not in self.undeclared:
                    logger.debug('possibly undeclared option: %s' % line )
                v = ' '.join(line[1:])
                if hasattr(self, k):
                    if type(getattr(self, k)) is float:
                        setattr(self, k, float(v))
                    elif type(getattr(self, k)) is int:
                        # the only integers that have units are durations.
                        # integers without units will come out unchanged.
                        setattr(self, k, durationToSeconds(v))
                    elif type(getattr(self, k)) is str:
                        setattr(self, k, [getattr(self, k), v])
                    elif type(getattr(self, k)) is list:
                        newv=getattr(self,k)
                        newv.append(v)
                        setattr(self, k, newv)
                else:
                    # FIXME:
                    setattr(self, k, v)
                    self.undeclared.append(k)

    def finalize(self, component=None, config=None):
        """ 
         Before final use, take the existing settings, and infer any missing needed defaults from what is provided.
         Should be called prior to using a configuration.

         There are default options that apply only if they are not overridden... 
       """

        self._parse_sum(None)

        if not component and self.component:
            component = self.component
            
        if not config and self.config:
            config = self.config
            
        if hasattr(self, 'nodupe_ttl'):
            if (type(self.nodupe_ttl) is str):
                if isTrue(self.nodupe_ttl):
                    self.nodupe_ttl = 300
                else:
                    self.nodupe_ttl = durationToSeconds(
                        self.nodupe_ttl, default=300)
        else:
            self.nodupe_ttl = 0

        if self.debug:
            self.logLevel = 'debug'

        if self.directory:
            self.directory = os.path.expanduser(self.directory)

        # double check to ensure duration options are properly parsed
        for d in duration_options:
            if hasattr(self, d) and (type(getattr(self, d)) is str):
                setattr(self, d, durationToSeconds(getattr(self, d)))

        if hasattr(self, 'kbytes_ps'):
            bytes_ps = humanfriendly.parse_size(self.kbytes_ps)
            if not self.kbytes_ps[-1].isalpha():
                bytes_ps *= 1024
            setattr(self, 'byteRateMax', bytes_ps)

        for d in count_options:
            if hasattr(self, d) and (type(getattr(self, d)) is str):
                setattr(self, d, humanfriendly.parse_size(getattr(self, d)))

        for d in size_options:
            if hasattr(self, d) and (type(getattr(self, d)) is str):
                setattr(self, d, chunksize_from_str(getattr(self, d)))

        for f in flag_options:
            if hasattr(self, f) and (type(getattr(self, f)) is str):
                setattr(self, f, isTrue(getattr(self, f)))

        for f in float_options:
            if hasattr(self, f) and (type(getattr(self, f)) is str):
                setattr(self, f, float(getattr(self, f)))

        if ( (len(self.logEvents) > 0 ) or self.log_flowcb_needed) :
            if ('sarracenia.flowcb.log.Log' not in self.plugins_late) and \
               ('log' not in self.plugins_late) :
                self.plugins_late.append( 'log' )

        # patch, as there is no 'none' level in python logging module...
        #    mapping so as not to break v2 configs.
        if hasattr(self, 'logLevel'):
            if self.logLevel == 'none':
                self.logLevel = 'critical'

        if hasattr(self, 'nodupe_basis'):
            if self.nodupe_basis == 'data': 
                self.plugins_early.append( 'nodupe.data' )
            elif self.nodupe_basis == 'name': 
                self.plugins_early.append( 'nodupe.name' )
            delattr( self, 'nodupe_basis' )

        # FIXME: note that v2 *user_cache_dir* is, v3 called:  cfg_run_dir
        if config[-5:] == '.conf':
            cfg = config[:-5]
        else:
            cfg = config

        if self.sanity_log_dead == 9999 :
            self.sanity_log_dead = 1.5*self.housekeeping
        if not hasattr(self, 'post_topicPrefix'):
           self.post_topicPrefix = self.topicPrefix

        if not hasattr(self, 'retry_ttl' ):
           self.retry_ttl = self.expire

        if self.retry_ttl == 0:
           self.retry_ttl = None

        if not hasattr(self, 'cfg_run_dir'):
            if self.statehost:
                hostdir = self.hostdir
            else:
                hostdir = None
            self.cfg_run_dir = os.path.join(get_user_cache_dir(hostdir),
                                            component, cfg)

        if self.post_broker is not None and self.post_broker.url is not None:
            if not hasattr(self,
                           'post_exchange') or self.post_exchange is None:
                self.post_exchange = 'xs_%s' % self.post_broker.url.username

            if hasattr(self, 'post_exchangeSuffix'):
                self.post_exchange += '_%s' % self.post_exchangeSuffix

            if hasattr(self,'post_exchange') and (type(self.post_exchange) is list ):
                pass
            elif hasattr(self, 'post_exchangeSplit') and self.post_exchangeSplit > 1:
                l = []
                for i in range(0, int(self.post_exchangeSplit)):
                    y = self.post_exchange + '%02d' % i
                    l.append(y)
                self.post_exchange = l
            else:
                self.post_exchange = [self.post_exchange]

            if (component in ['poll' ]) and (hasattr(self,'vip') and self.vip):
                if (not hasattr(self,'exchange') or not self.exchange):
                    if type(self.post_exchange) is list:
                        self.exchange = self.post_exchange[0]
                    else:
                        self.exchange = self.post_exchange
                if (not hasattr(self,'broker') or not self.broker):
                    self.broker = self.post_broker


        if self.sourceFromExchange and self.exchange:
           self.source = self.get_source_from_exchange(self.exchange)

        if self.broker and self.broker.url and self.broker.url.username:
            self._resolve_exchange()

            queuefile = sarracenia.user_cache_dir(
                Config.appdir_stuff['appname'],
                Config.appdir_stuff['appauthor'])

            if self.statehost:
                queuefile += os.sep + self.hostdir

            queuefile += os.sep + component + os.sep + cfg
            queuefile += os.sep + component + '.' + cfg + '.' + self.broker.url.username

            if hasattr(self, 'exchangeSplit') and hasattr(
                    self, 'no') and (self.no > 0):
                queuefile += "%02d" % self.no
            queuefile += '.qname'

            self.queue_filename = queuefile

            #while (not hasattr(self, 'queueName')) or (self.queueName is None):
            """

              normal:
                  if not the lead instance, wait a bit for the queuefile to be written.
                  look for a queuefile in the state directory, if it is there, read it.
                  if you can't read the file

              if you are instance 1, or 0 (foreground) and the queuefile is missing, then need
              to write it.  if queueName is set, use that, if not

              if you set the queuename, it might have variable values that when evaluated repeatedly (such as randomized settings)
              will come out differently every time. So even in the case of a fixed queue name, need to write 

            """

            if hasattr(self,'no') and self.no > 1:
                # worker instances need give lead instance time to write the queuefile
                time.sleep(randint(4,14))

                queue_file_read=False
                config_read_try=0
                while not queue_file_read:
                    if os.path.isfile(queuefile):
                        f = open(queuefile, 'r')
                        self.queueName = f.read()
                        f.close()
                    else:
                        self.queueName = ''

                    config_read_try += 1
                    logger.debug( f'instance read try {config_read_try} queueName {self.queueName} from queue state file {queuefile}' )
                    if len(self.queueName) < 1:
                          nap=randint(1,4)
                          logger.debug( f'queue name corrupt take a short {nap} second nap, then try again' )
                          time.sleep(nap)
                          if config_read_try > 5:
                              logger.critical( f'failed to read queue name from {queuefile}')
                              sys.exit(2)
                    else:
                          queue_file_read=True

            else: # lead instance just tries once...

                if os.path.isfile(queuefile):
                    f = open(queuefile, 'r')
                    self.queueName = f.read()
                    f.close()
                
            #if the queuefile is corrupt, then will need to guess anyways.
            if ( self.queueName is None ) or ( self.queueName == '' ):
                queueName = 'q_' + self.broker.url.username + '_' + component + '.' + cfg
                if hasattr(self, 'queue_suffix'):
                    queueName += '.' + self.queue_suffix
                queueName += '.' + str(randint(0, 100000000)).zfill(8)
                queueName += '.' + str(randint(0, 100000000)).zfill(8)
                self.queueName = queueName
                logger.debug('default guessed queueName  %s ' % self.queueName )

            if not os.path.isdir(os.path.dirname(queuefile)):
                pathlib.Path(os.path.dirname(queuefile)).mkdir(parents=True, exist_ok=True)

            # only lead instance (0-foreground, 1-start, or none in the case of 'declare')
            # should write the state file.
            if (self.queueName is not None) and (not hasattr(self,'no') or (self.no < 2)):
                f = open(queuefile, 'w')
                f.write(self.queueName)
                f.close()
                logger.debug( f'queue name {self.queueName} persisted to {queuefile}' )

        if hasattr(self, 'no'):
            if self.statehost:
                hostdir = self.hostdir
            else:
                hostdir = None
            self.metricsFilename = get_metrics_filename(hostdir, component, cfg, self.no)
            self.pid_filename = get_pid_filename(hostdir, component, cfg, self.no)
            self.retry_path = self.pid_filename.replace('.pid', '.retry')
            self.novipFilename = self.pid_filename.replace('.pid', '.noVip')


        if (self.bindings == [] and hasattr(self, 'exchange')):
            self.bindings = [(self.exchange, self.topicPrefix, [ '#' ])]

        if hasattr(self, 'documentRoot') and (self.documentRoot is not None):
            path = os.path.expanduser(os.path.abspath(self.documentRoot))
            if self.realpathPost:
                path = os.path.realpath(path)

            if sys.platform == 'win32' and words0.find('\\'):
                logger.warning("%s %s" % (words0, words1))
                logger.warning(
                    "use of backslash ( \\ ) is an escape character. For a path separator use forward slash ( / )."
                )

            if sys.platform == 'win32':
                self.documentRoot = path.replace('\\', '/')
            else:
                self.documentRoot = path
            n = 2

        if hasattr(self, 'pollUrl'):
            if not hasattr(self,'post_baseUrl') or not self.post_baseUrl :
                logger.debug( f"defaulting post_baseUrl to match pollURl, since it isn't specified." )
                self.post_baseUrl = self.pollUrl
            
        # verify post_baseDir

        if self.post_baseDir is None:
            if self.post_documentRoot is not None:
                self.post_baseDir = os.path.expanduser(self.post_documentRoot)
                logger.warning("use post_baseDir instead of post_documentRoot")
            elif self.documentRoot is not None:
                self.post_baseDir = os.path.expanduser(self.documentRoot)
                logger.warning("use post_baseDir instead of documentRoot")
            elif self.post_baseUrl and ( self.post_baseUrl[0:5] in [ 'file:' ] ):
                self.post_baseDir = self.post_baseUrl[5:]
            elif self.post_baseUrl and ( self.post_baseUrl[0:5] in [ 'sftp:' ] ):
                u =  sarracenia.baseUrlParse(self.post_baseUrl) 
                self.post_baseDir = u.path
            elif self.baseDir is not None:
                self.post_baseDir = os.path.expanduser(self.baseDir)
                logger.debug("defaulting post_baseDir to same as baseDir")


        if self.messageCountMax > 0:
            if self.batch > self.messageCountMax:
                self.batch = self.messageCountMax
                logger.info(
                    'overriding batch for consistency with messageCountMax: %d'
                    % self.batch)

        if (component not in ['poll' ]):
            self.path = list(map( os.path.expanduser, self.path ))

        if self.vip and not features['vip']['present']:
            logger.critical( f"vip feature requested, but missing library: {' '.join(features['vip']['modules_needed'])} " )
            sys.exit(1)

    def check_undeclared_options(self):

        alloptions = str_options + flag_options + float_options + list_options + set_options + count_options + size_options + duration_options
        # FIXME: confused about this...  commenting out for now...
        for u in self.undeclared:
            if u not in alloptions:
                logger.error("undeclared option: %s" % u)

        no_defaults=set()
        for u in alloptions:
             if not hasattr(self,u):
                no_defaults.add( u )

        logger.debug("missing defaults: %s" % no_defaults)

    """
      2020/05/26 FIXME here begins sheer terror.
      following routines are taken verbatim from v2. 
      trying not to touch it... it is painful.
      setting new_ values for downloading etc...
      sundew_* ... 
   """

    def _sundew_basename_parts(self, pattern, basename):
        """
        modified from metpx SenderFTP
        """

        if pattern == None: return []
        parts = re.findall(pattern, basename)
        if len(parts) == 2 and parts[1] == '': parts.pop(1)
        if len(parts) != 1: return None

        lst = []
        if isinstance(parts[0], tuple):
            lst = list(parts[0])
        else:
            lst.append(parts[0])

        return lst

    # from metpx SenderFTP
    def sundew_dirPattern(self, pattern, urlstr, basename, destDir):
        """
        does substitutions for patterns in directories.

        """
        BN = basename.split(":")
        EN = BN[0].split("_")

        BP = self._sundew_basename_parts(pattern, urlstr)

        ndestDir = ""
        DD = destDir.split("/")
        for ddword in DD:
            if ddword == "": continue

            nddword = ""
            DW = ddword.split("$")
            for dwword in DW:
                nddword += self.sundew_matchPattern(BN, EN, BP, dwword, dwword)

            ndestDir += "/" + nddword

        # This code might add an unwanted '/' in front of ndestDir
        # if destDir does not start with a substitution $ and
        # if destDir does not start with a / ... it does not need one

        if (len(destDir) > 0) and (destDir[0] != '$') and (destDir[0] != '/'):
            if ndestDir[0] == '/': ndestDir = ndestDir[1:]

        return ndestDir

    # modified from metpx SenderFTP
    def sundew_matchPattern(self, BN, EN, BP, keywd, defval):

        BN6 = time.strftime("%Y%m%d%H%M%S", time.gmtime())
        if len(BN) >= 7: BN6 = BN[6]

        if keywd[:4] == "{T1}": return (EN[0])[0:1] + keywd[4:]
        elif keywd[:4] == "{T2}": return (EN[0])[1:2] + keywd[4:]
        elif keywd[:4] == "{A1}": return (EN[0])[2:3] + keywd[4:]
        elif keywd[:4] == "{A2}": return (EN[0])[3:4] + keywd[4:]
        elif keywd[:4] == "{ii}": return (EN[0])[4:6] + keywd[4:]
        elif keywd[:6] == "{CCCC}": return EN[1] + keywd[6:]
        elif keywd[:4] == "{YY}": return (EN[2])[0:2] + keywd[4:]
        elif keywd[:4] == "{GG}": return (EN[2])[2:4] + keywd[4:]
        elif keywd[:4] == "{Gg}": return (EN[2])[4:6] + keywd[4:]
        elif keywd[:5] == "{BBB}":
            return (EN[3])[0:3] + keywd[5:]
            # from pds'datetime suffix... not sure
        elif keywd[:7] == "{RYYYY}":
            return BN6[0:4] + keywd[7:]
        elif keywd[:5] == "{RMM}":
            return BN6[4:6] + keywd[5:]
        elif keywd[:5] == "{RDD}":
            return BN6[6:8] + keywd[5:]
        elif keywd[:5] == "{RHH}":
            return BN6[8:10] + keywd[5:]
        elif keywd[:5] == "{RMN}":
            return BN6[10:12] + keywd[5:]
        elif keywd[:5] == "{RSS}":
            return BN6[12:14] + keywd[5:]

        # Matching with basename parts if given

        if BP != None:
            for i, v in enumerate(BP):
                kw = '{' + str(i) + '}'
                lkw = len(kw)
                if keywd[:lkw] == kw: return v + keywd[lkw:]

        return defval

    def variableExpansion(self, cdir, message=None ):
        """
            replace substitution patterns, variable substitutions as described in
            https://metpx.github.io/sarracenia/Reference/sr3_options.7.html#variables

            returns:  the given string with the substiturions done.

            examples:   ${YYYYMMDD-70m} becomes 20221107 assuming that was the current date 70 minutes ago.
                        environment variables, and built-in settings are replaced also.

           timeoffset -70m


        """

        if not '$' in cdir:
            return cdir

        new_dir = cdir

        while '${BD}' in new_dir and self.baseDir != None:
            new_dir = new_dir.replace('${BD}', self.baseDir, 1)

        while ( '${BUP}' in new_dir ) and ( 'baseUrl' in message ):
            u = sarracenia.baseUrlParse( message['baseUrl'] )
            new_dir = new_dir.replace('${BUP}', u.path, 1 )

        while ( '${baseUrlPath}' in new_dir ) and ( 'baseUrl' in message ):
            u = sarracenia.baseUrlParse( message['baseUrl'] )
            new_dir = new_dir.replace('${baseUrlPath}', u.path, 1)

        while ( '${BUPL}' in new_dir ) and ( 'baseUrl' in message ):
            u = sarracenia.baseUrlParse( message['baseUrl'] )
            new_dir = new_dir.replace('${BUPL}', os.path.basename(u.path), 1 )

        while ( '${baseUrlPathLast}' in new_dir )  and ( 'baseUrl' in message ):
            u = sarracenia.baseUrlParse( message['baseUrl'] )
            new_dir = new_dir.replace('${baseUrlPathLast}', os.path.basename(u.path), 1 )

        while '${PBD}' in new_dir and self.post_baseDir != None:
            new_dir = new_dir.replace('${PBD}', self.post_baseDir, 1)

        while '${DR}' in new_dir and self.documentRoot != None:
            logger.warning(
                "DR = documentRoot should be replaced by BD for base_dir")
            new_dir = new_dir.replace('${DR}', self.documentRoot, 1)

        while '${PDR}' in new_dir and self.post_baseDir != None:
            logger.warning(
                "PDR = post_documentRoot should be replaced by PBD for post_baseDir"
            )
            new_dir = new_dir.replace('${PDR}', self.post_baseDir, 1)

        #whenStamp = time.gmtime( time.time()+self.varTimeOffset )

        whenStamp = datetime.datetime.fromtimestamp( time.time()+self.varTimeOffset )

        while '${YYYYMMDD}' in new_dir:
            YYYYMMDD = whenStamp.strftime("%Y%m%d")
            new_dir = new_dir.replace('${YYYYMMDD}', YYYYMMDD)

        while '${SOURCE}' in new_dir:
            new_dir = new_dir.replace('${SOURCE}', message['source'])

        while '${DD}' in new_dir:
            DD = whenStamp.strftime("%d")
            new_dir = new_dir.replace('${DD}', DD)

        while '${HH}' in new_dir:
            HH = whenStamp.strftime("%H")
            new_dir = new_dir.replace('${HH}', HH)

        while '${YYYY}' in new_dir:
            YYYY = whenStamp.strftime("%Y")
            new_dir = new_dir.replace('${YYYY}', YYYY)

        while '${MM}' in new_dir:
            MM = whenStamp.strftime("%m")
            new_dir = new_dir.replace('${MM}', MM)

        while '${JJJ}' in new_dir:
            JJJ = whenStamp.strftime("%j")
            new_dir = new_dir.replace('${JJJ}', JJJ)

        # strftime compatible patterns.
        fragment_list=[]
        for fragment in new_dir.split( '${%' ):
            close_brace = fragment.find('}')
            frag_start=0
            seconds=self.varTimeOffset

            # only support %o time offsets at the beginning of the string.
            if fragment[0] in ['+','-','o', '0', '1', '2', '3', '4', '5','6','7', '8',  '9' ]:
                end_of_offset=fragment.find('%')
                if fragment[0] == 'o':
                    s= 2 if fragment[1] in [ '-','+' ] else 1
                else:
                    s= 1 if fragment[0] in [ '-','+' ] else 0
                seconds = durationToSeconds(fragment[s:end_of_offset])
                frag_start=end_of_offset+1
                if '-' in fragment[0:2]: 
                    seconds = -1 * seconds

            whenStamp = datetime.datetime.fromtimestamp( time.time()+seconds )
        
            if close_brace > 0:
                time_str=whenStamp.strftime( "%"+fragment[frag_start:close_brace] )
                fragment_list.append(time_str)
                fragment_list.append(fragment[close_brace+1:])
            else:
                fragment_list.append(fragment)
        new_dir=''.join(fragment_list)
        
        # Parsing cdir to subtract time from it in the following formats
        # time unit can be: sec/mins/hours/days/weeks

        # ${YYYY-[number][time_unit]}
        offset_check = re.search( r'\$\{YYYY-(\d+)(\D)\}', cdir)
        if offset_check:
            logger.info( f"offset 0: {offset_check.group(1,2)}" )
            seconds = durationToSeconds(''.join(offset_check.group(1, 2)),
                                             's')

            epoch = time.mktime(time.gmtime()) - seconds
            YYYY1D = time.strftime("%Y", time.localtime(epoch))
            new_dir = re.sub( r'\$\{YYYY-\d+\D\}', YYYY1D, new_dir)

        # ${MM-[number][time_unit]}
        offset_check = re.search( r'\$\{MM-(\d+)(\D)\}', cdir)
        if offset_check:
            logger.info( f"offset 1: {offset_check.group(1,2)}" )
            seconds = durationToSeconds(''.join(offset_check.group(1, 2)),
                                             's')

            epoch = time.mktime(time.gmtime()) - seconds
            MM1D = time.strftime("%m", time.localtime(epoch))
            new_dir = re.sub( r'\$\{MM-\d+\D\}', MM1D, new_dir)

        # ${JJJ-[number][time_unit]}
        offset_check = re.search(r'\$\{JJJ-(\d+)(\D)\}', cdir)
        if offset_check:
            logger.info( f"offset 2: {offset_check.group(1,2)}" )
            seconds = durationToSeconds(''.join(offset_check.group(1, 2)),
                                             's')

            epoch = time.mktime(time.gmtime()) - seconds
            JJJ1D = time.strftime("%j", time.localtime(epoch))
            new_dir = re.sub( r'\$\{JJJ-\d+\D\}', JJJ1D, new_dir)

        # ${YYYYMMDD-[number][time_unit]}
        offset_check = re.search(r'\$\{YYYYMMDD-(\d+)(\D)\}', cdir)
        if offset_check:
            logger.info( f"offset 3: {offset_check.group(1,2)}" )
            seconds = durationToSeconds(''.join(offset_check.group(1, 2)),
                                             's')
            epoch = time.mktime(time.gmtime()) - seconds
            YYYYMMDD = time.strftime("%Y%m%d", time.localtime(epoch))
            logger.info( f"seconds: {seconds} YYYYMMDD {YYYYMMDD}" )
            new_dir = re.sub( r'\$\{YYYYMMDD-\d+\D\}', YYYYMMDD, new_dir)

        new_dir = self._varsub(new_dir)

        return new_dir


    """
       2020/05/26 PAS... FIXME: end of sheer terror. 

       the parts below used be part of the sheer terror... but have been
       tamed a bit.
    """

    class addBinding(argparse.Action):
        """
        called by argparse to deal with queue bindings.
        """
        def __call__(self, parser, namespace, values, option_string):

            if values == 'None':
                namespace.bindings = []

            namespace._resolve_exchange()

            if not hasattr(namespace, 'broker'):
                raise Exception('broker needed before subtopic')
                return

            if not hasattr(namespace, 'exchange'):
                raise Exception('exchange needed before subtopic')
                return

            if not hasattr(namespace, 'topicPrefix'):
                raise Exception('topicPrefix needed before subtopic')
                return

            if type(namespace.topicPrefix) is str:
               if namespace.broker.scheme[0:3] == 'amq':
                   topicPrefix = namespace.topicPrefix.split('.')
               else:
                   topicPrefix = namespace.topicPrefix.split('/')

            namespace.bindings.append(
                (namespace.exchange, topicPrefix, values))

    def parse_args(self, isPost=False):
        """
        user information:
           accept a configguration, apply argParse library to augment the given configuration
           with command line settings.

           the post component has a different calling convention than others, so use that flag
           if called from post.

        development notes:
           Use argparse.parser to modify defaults.
           FIXME, many FIXME notes below. this is a currently unusable placeholder.
           have not figured this out yet. many issues.

           FIXME #1:
           parseArgs often sets the value of the variable, regardless of it's presence (normally a good thing.)
           ( if you have 'store_true' then default needed, for broker, just a string, it ignores if not present.)
           This has the effect of overriding settings in the file parsed before the arguments.
           Therefore: often supply defaults... but... sigh...
           
           but there is another consideration stopping me from supplying defaults, wish I remembered what it was.
           I think it is:
           FIXME #2: 
           arguments are parsed twice: once to get basic stuff (loglevel, component, action)
           and if the parsing fails there, the usage will print the wrong defaults... 

        """

        parser=argparse.ArgumentParser( \
             description='version: %s\nSarracenia flexible tree copy ( https://MetPX.github.io/sarracenia ) ' % sarracenia.__version__ ,\
             formatter_class=argparse.ArgumentDefaultsHelpFormatter )

        if sys.version_info[0] >= 3 and sys.version_info[1] < 8:
            parser.register('action', 'extend', ExtendAction)
        
        parser.add_argument('--acceptUnmatched',
                            default=self.acceptUnmatched,
                            type=bool,
                            nargs='?',
                            help='default selection, if nothing matches')
        parser.add_argument(
            '--action',
            '-a',
            nargs='?',
            choices=Config.actions,
            help='action to take on the specified configurations')
        parser.add_argument('--admin',
                            help='amqp://user@host of peer to manage')
        parser.add_argument(
            '--attempts',
            type=int,
            nargs='?',
            help='how many times to try before queuing for retry')
        parser.add_argument(
            '--base_dir',
            '-bd',
            nargs='?',
            help="path to root of tree for relPaths in messages.")
        parser.add_argument('--batch',
                            type=int,
                            nargs='?',
                            help='how many transfers per each connection')
        parser.add_argument(
            '--blocksize',
            type=int,
            nargs='?',
            help=
            'size to partition files. 0-guess, 1-never, any other number: that size'
        )
        """
           FIXME:  Most of this is gobblygook place holder stuff, by copying from wmo-mesh example.
           Don't really need this to work right now, so just leaving it around as-is.  Challenges:

           -- sizing units,  K, M, G,  (should have humanfriendly based parsing.)
           -- time units s,h,m,d 
           -- what to do with verbs.
           -- accept/reject whole mess requires extension deriving a class from argparse.Action.
           
        """
        parser.add_argument('--broker',
                            nargs='?',
                            help='amqp://user:pw@host of peer to subscribe to')
        parser.add_argument('--config',
                            '-c',
                            nargs='?',
                            help=' specifical configuration to select ')
        parser.add_argument('--dangerWillRobinson',
                            type=int,
                            default=0,
                            help='Confirm you want to do something dangerous')
        parser.add_argument('--debug',
                            action='store_true',
                            default=self.debug,
                            help='print debugging output (very verbose)')
        parser.add_argument('--dry_run', '--simulate', '--simulation', 
                            action='store_true',
                            default=self.dry_run,
                            help='simulation mode (perform no file transfers, just print what would happen)')
        parser.add_argument('--exchange',
                            nargs='?',
                            default=self.exchange,
                            help='root of the topic tree to subscribe to')

        parser.add_argument('--full',
                            action='store_true',
                            default=self.displayFull,
                            help='fuller, more verbose display')
        """
        FIXME: header option not implemented in argparsing: should add to the fixed_header dictionary.
          
        """
        """
        FIXME: in previous parser, exchange is a modifier for bindings, can have several different values for different subtopic bindings.
           as currently coded, just a single value that over-writes previous setting, so only binding to a single exchange is possible.
        """

        parser.add_argument('--inline',
                            dest='inline',
                            default=self.inline,
                            action='store_true',
                            help='include file data in the message')
        parser.add_argument(
            '--inlineEncoding',
            choices=['text', 'binary', 'guess'],
            default=self.inlineEncoding,
            help='encode payload in base64 (for binary) or text (utf-8)')
        parser.add_argument('--inlineByteMax',
                            type=int,
                            default=self.inlineByteMax,
                            help='maximum message size to inline')
        parser.add_argument(
            '--instances',
            type=int,
            help='number of processes to run per configuration')

        parser.add_argument('--identity_method', '--identity', '-s', '--sum',
                            nargs='?',
                            default=self.identity_method,
                            help='choose a different checksumming method for the files posted')
        if hasattr(self, 'bindings'):
            parser.set_defaults(bindings=self.bindings)


        parser.add_argument(
            '--logLevel',
            choices=[
                'notset', 'debug', 'info', 'warning', 'error', 'critical'
            ],
            help='encode payload in base64 (for binary) or text (utf-8)')
        parser.add_argument('--logReject',
                            action='store_true',
                            default=self.logReject,
                            help='print a log message explaining why each file is rejected')
        parser.add_argument('--logStdout',
                            action='store_true',
                            default=False,
                            help='disable logging, everything to standard output/error')
        parser.add_argument('--no',
                            type=int,
                            help='instance number of this process')
        parser.add_argument('--queueName',
                            nargs='?',
                            help='name of AMQP consumer queue to create')
        parser.add_argument('--post_broker',
                            nargs='?',
                            help='broker to post downloaded files to')
        #parser.add_argument('--post_baseUrl', help='base url of the files announced')
        parser.add_argument('--post_exchange',
                            nargs='?',
                            help='root of the topic tree to announce')
        parser.add_argument(
            '--post_exchangeSplit',
            type=int,
            nargs='?',
            help='split output into different exchanges 00,01,...')
        parser.add_argument(
            '--post_topicPrefix',
            nargs='?',
            help=
            'allows simultaneous use of multiple versions and types of messages'
        )
        #FIXME: select/accept/reject in parser not implemented.
        parser.add_argument(
            '--select',
            nargs=1,
            action='append',
            help='client-side filtering: accept/reject <regexp>')
        parser.add_argument(
            '--subtopic',
            nargs=1,
            action=Config.addBinding,
            help=
            'server-side filtering: MQTT subtopic, wilcards # to match rest, + to match one topic'
        )
        parser.add_argument(
            '--topicPrefix',
            nargs='?',
            default=self.topicPrefix,
            help=
            'allows simultaneous use of multiple versions and types of messages'
        )
        parser.add_argument('--users',
                            default=False,
                            action='store_true',
                            help='only for declare... declare users?')
        parser.add_argument(
            '--version',
            '-v',
            action='version',
            version='%s' % sarracenia.__version__,
            help=
            'server-side filtering: MQTT subtopic, wilcards # to match rest, + to match one topic'
        )

        if isPost:
            parser.add_argument('--path',
                                '-p',
                                action='append',
                                nargs='?',
                                help='path to post or watch')
            parser.add_argument('path',
                                nargs='*',
                                action='extend',
                                help='files to post')
        else:
            parser.add_argument(
                'action',
                nargs='?',
                choices=Config.actions,
                help='action to take on the specified configurations')
            parser.add_argument('configurations',
                                nargs='*',
                                help='configurations to operate on')

        args = parser.parse_args()

        if hasattr(args, 'help'):
            args.print_usage()

        if hasattr(args, 'config') and (args.config is not None):
            args.configurations = [args.config]

        if hasattr(args,'full'):
            self.displayFull = args.full
            delattr(args,'full')

        self.merge(args)


def default_config():

    cfg = Config()
    cfg.currentDir = None
    cfg.override(default_options)
    cfg.override(sarracenia.moth.default_options)
    if features['amqp']['present']:
        cfg.override(sarracenia.moth.amqp.default_options)
    cfg.override(sarracenia.flow.default_options)

    for g in ["admin.conf", "default.conf"]:
        if os.path.exists(get_user_config_dir() + os.sep + g):
            cfg.parse_file(get_user_config_dir() + os.sep + g)

    return cfg

def no_file_config():
    """
      initialize a config that will not use Sarracenia configuration files at all.
      meant for use by people writing independent programs to start up instances
      with python API calls.

    """
    cfg = Config()
    cfg.currentDir = None
    cfg.override(default_options)
    cfg.override(sarracenia.moth.default_options)
    if features['amqp']['present']:
        cfg.override(sarracenia.moth.amqp.default_options)
    cfg.override(sarracenia.flow.default_options)
    cfg.cfg_run_dir = '.'
    cfg.retry_path = '.'
    return cfg


def one_config(component, config, isPost=False):
    """
      single call return a fully parsed single configuration for a single component to run.

      read in admin.conf and default.conf

      apply component default overrides ( maps to: component/check ?)
      read in component/config.conf
      parse arguments from command line.
      return config instance item.

      
      appdir_stuff can be to override file locations for testing during development.

    """
    default_cfg = default_config()
    #default_cfg.override(  { 'component':component, 'directory': os.getcwd(), 'acceptUnmatched':True, 'no':0 } )
    default_cfg.override({
        'component': component,
        'config': config,
        'acceptUnmatched': True,
        'no': 0
    })

    cfg = copy.deepcopy(default_cfg)

    cfg.applyComponentDefaults( component )

    store_pwd = os.getcwd()

    os.chdir(get_user_config_dir())
    os.chdir(component)

    if config[-5:] != '.conf':
        fname = os.path.expanduser(config + '.conf')
    else:
        fname = os.path.expanduser(config)

    if os.path.exists(fname):
         cfg.parse_file(fname,component)
    else:
         logger.error('config %s not found' % fname )
         return None

    os.chdir(store_pwd)

    cfg.parse_args(isPost)

    #logger.error( 'after args' )
    #print( 'after args' )
    #cfg.dump()
    if component in ['poll' ]:
        if not hasattr(cfg,'broker') or (cfg.broker is None):
             cfg.broker = cfg.post_broker

    cfg.finalize(component, config)

    if component in ['post', 'watch']:
        cfg.postpath = list( map( os.path.expanduser, cfg.configurations[1:]))
        if hasattr(cfg, 'path') and (cfg is not None):
            if type(cfg.path) is list:
                cfg.postpath.extend(cfg.path)
            else:
                cfg.postpath.append(cfg.path)
            logger.debug('path is : %s' % cfg.path)
            logger.debug('postpath is : %s' % cfg.postpath)
        
    #pp = pprint.PrettyPrinter(depth=6)
    #pp.pprint(cfg)

    return cfg

def cfglogs(cfg_preparse, component, config, logLevel, child_inst):


    if cfg_preparse.logRotateInterval < 24*24*60:
        logRotateInterval=int(cfg_preparse.logRotateInterval)
        lr_when='s'
    else:
        logRotateInterval = int(cfg_preparse.logRotateInterval/(24*24*60))
        lr_when='midnight'

    # init logs here. need to know instance number and configuration and component before here.

    if cfg_preparse.action == 'start' and not cfg_preparse.logStdout:
        if cfg_preparse.statehost:
            hostdir = cfg_preparse.hostdir
        else:
            hostdir = None

        metricsfilename = get_metrics_filename( hostdir, component, config, child_inst)

        dir_not_there = not os.path.exists(os.path.dirname(metricsfilename))
        while dir_not_there:
            try:
                os.makedirs(os.path.dirname(metricsfilename), exist_ok=True)
                dir_not_there = False
            except FileExistsError:
                dir_not_there = False
            except Exception as ex:
                logging.error( "makedirs {} failed err={}".format(os.path.dirname(metricsfilename),ex))
                logging.debug("Exception details:", exc_info=True)
                os.sleep(1)

        cfg_preparse.metricsFilename = metricsfilename

        logfilename = get_log_filename( hostdir, component, config, child_inst)

        dir_not_there = not os.path.exists(os.path.dirname(logfilename))
        while dir_not_there:
            try:
                os.makedirs(os.path.dirname(logfilename), exist_ok=True)
                dir_not_there = False
            except FileExistsError:
                dir_not_there = False
            except Exception as ex:
                logging.error( "makedirs {} failed err={}".format(os.path.dirname(logfilename),ex))
                logging.debug("Exception details:", exc_info=True)
                os.sleep(1)

        log_format = '%(asctime)s [%(levelname)s] %(name)s %(funcName)s %(message)s'
        if logging.getLogger().hasHandlers():
            for h in logging.getLogger().handlers:
                h.close()
                logging.getLogger().removeHandler(h)
        logger = logging.getLogger()
        logger.setLevel(logLevel.upper())

        handler = sarracenia.instance.RedirectedTimedRotatingFileHandler(
            logfilename,
            when=lr_when,
            interval=logRotateInterval,
            backupCount=cfg_preparse.logRotateCount)
        handler.setFormatter(logging.Formatter(log_format))

        logger.addHandler(handler)

        if hasattr(cfg_preparse, 'permLog'):
            os.chmod(logfilename, cfg_preparse.permLog)

        # FIXME: https://docs.python.org/3/library/contextlib.html portable redirection...
        if sys.platform != 'win32':
            os.dup2(handler.stream.fileno(), 1)
            os.dup2(handler.stream.fileno(), 2)

    else:
        try:
            logger.setLevel(logLevel)
        except Exception:
            logger.setLevel(logging.INFO)

# add directory to python front of search path for plugins.
plugin_dir = get_user_config_dir() + os.sep + "plugins"
if os.path.isdir(plugin_dir) and not plugin_dir in sys.path:
    sys.path.insert(0, plugin_dir)



import json
import logging
import os.path
import re

from sarracenia.flowcb import FlowCB
import GTStoWIS2
import uuid

logger = logging.getLogger(__name__)


class Wistree(FlowCB):
    """
       Given a file whose name begins with a WMO GTS AHL
       (World Meteorological Organization, Global Telecommunications' System, Abbreviated Header Line)
       map the given AHL to a WIS-compliant topic tree. (WIS-WMO Information Service.)

       So when downloading, instead of writing the file to a single directory, it is
       written to a WIS-compliant folder structure.
    """
    def __init__(self, options):

        super().__init__(options,logger)
        self.topic_builder = GTStoWIS2.GTStoWIS2()
        self.date_pattern = re.compile("^[0-9]{8}$")

    def after_accept(self, worklist):

        new_incoming = []

        for msg in worklist.incoming:

            # /20181218/UCAR-UNIDATA/WMO-BULLETINS/IX/21/IXTD99_KNES_182147_9d73fc80e12fca52a06bf41c716cd718

            try:
                # fix file name suffix.
                type_suffix = self.topic_builder.mapAHLtoExtension(
                    msg['new_file'][0:2])

                # /20181218/UCAR-UNIDATA/WMO-BULLETINS/IX/21/IXTD99_KNES_182147_9d73fc80e12fca52a06bf41c716cd718.cap
                tpfx = msg['subtopic']

                msg['id']= str(uuid.uuid4())

                # input has relpath=/YYYYMMDD/... + pubTime
                # need to move the date from relPath to BaseDir, adding the T hour from pubTime.
                if self.date_pattern.match(tpfx[0]):
                    new_baseSubDir = tpfx[0] + msg['pubTime'][8:11]
                else:
                    # or default to using pubTime...
                    new_baseSubDir = msg['pubTime'][0:11]


                new_baseDir = msg['new_dir'] + os.sep + new_baseSubDir
                new_relDir = str(uuid.uuid4()).replace('-','/')
                msg['topic'] = 'WIS' + os.sep + self.topic_builder.mapAHLtoTopic( msg['new_file'])
                msg['new_file'] = str(uuid.uuid4())

                if msg['new_file'][-len(type_suffix):] != type_suffix:
                    new_file = msg['new_file'] + type_suffix
                else:
                    new_file = msg['new_file']

                if type_suffix == 'bufr' :
                    mtype='application/x-bufr'
                elif type_suffix == 'grib':
                    mtype='application/x-grib'
                else:
                    mtype='unknown'

                msg['links'] =  { "href": msg['baseUrl'] + '/' + msg['relPath'], 'rel':'canonical', 'type': mtype }
                msg.updatePaths(self.o, new_baseDir + os.sep + new_relDir,
                                new_file)

            except Exception as ex:
                logger.error("failed to map %s to a topic, skipped." %
                             msg['new_file'],
                             exc_info=True)
                worklist.failed.append(msg)
                continue

            # remove some legacy fields in the messages, to de-clutter.
            msg['_deleteOnPost'] |= set([
                'from_cluster',
                'sum',
                'to_clusters',
                'sundew_extension',
                'filename',
                'source',
            ])
            new_incoming.append(msg)

        worklist.incoming = new_incoming




import logging

from sarracenia.flowcb import FlowCB
from sarracenia import timestr2flt, nowflt

logger = logging.getLogger(__name__)


class MDelayLatest(FlowCB):
    """
      This plugin delays processing of messages by *message_delay* seconds
      If multiple versions of a file are published within the interval, only the latest one
      will be published.


      mdelay 30
      flowcb sarracenia.flowcb.mdelaylatest.MDelayLatest

      every message will be at least 30 seconds old before it is forwarded by this plugin.
      In the meantime, the message is placed on the retry queue by marking it as failed.

    """
    def __init__(self, options):

        super().__init__(options,logger)
        self.stop_requested = False
        self.suppressions = 0
        self.ok_delay = []
        logging.basicConfig(format=self.o.logFormat,
                            level=getattr(logging, self.o.logLevel.upper()))

        self.o.add_option('mdelay', 'duration', 30)

        logger.info(f'mdelay set to {self.o.mdelay}')

    def after_accept(self, worklist):

        # Check message in ok list
        # get time at beginning of loop, less system calls.
        now = nowflt()

        new_incoming = []
        for m1 in worklist.incoming:
            #logger.info('1 relPath=%s' %  m1['relPath'])
            #logger.info('1 pubTime=%s' %  m1['pubTime'])
            elapsedtime = now - timestr2flt(m1['pubTime'])
            #logger.info('1 Time=%s' %  str(elapsedtime))
            wait = False

            # If same message found in the delay list, replaced it with the one in ok list.
            new_ok_delay = []
            for m2 in self.ok_delay:
                if m1['relPath'] == m2['relPath']:
                    logger.info(
                        f"intermediate version suppressed: {m1['relPath']}")
                    self.suppressions += 1
                    new_ok_delay.append(m1)
                    worklist.rejected.append(m2)
                    wait = True
                else:
                    new_ok_delay.append(m2)
                    #new_incoming.append(m1)
            self.ok_delay = new_ok_delay

            # If it's new, put it in delay list too.
            # else if replacing a message in delay, do nothing
            # else put it back to incoming
            if not wait and elapsedtime < self.o.mdelay:
                self.ok_delay.append(m1)
            elif wait:
                pass
            else:
                new_incoming.append(m1)

        worklist.incoming = new_incoming

        # Check message in the delay list
        new_ok_delay = []
        for m1 in self.ok_delay:
            #logger.info('2 relPath=%s' %  m1['relPath'])
            #logger.info('2 pubTime=%s' %  m1['pubTime'])
            elapsedtime = nowflt() - timestr2flt(m1['pubTime'])
            #logger.info('Time=%s' %  str(elapsedtime))
            # if it's time, the message is putting back to the ok list to publish
            if elapsedtime >= self.o.mdelay:
                #logger.info('OK')
                worklist.incoming.append(m1)
            else:
                new_ok_delay.append(m1)
        self.ok_delay = new_ok_delay

        if self.stop_requested:
            worklist.incoming.extend(self.ok_delay)
            self.ok_delay = []
            logger.info('stop requested, ok_delay queue drained')

    def on_housekeeping(self):
        logger.info(
            f'suppressions={self.suppressions} currently delay queue length:{len(self.ok_delay)}'
        )
        self.suppressions = 0




from sarracenia.flow import Flow
import logging

logger = logging.getLogger(__name__)

default_options = {'acceptUnmatched': True, 'nodupe_ttl': 0}


class Report(Flow):
    """
      forward report messages. 

      Not really implemented at the moment. It is just a shovel synonym for now.
      more logic should be added.
    """
    def __init__(self, options):

        super().__init__(options)
        self.plugins['load'].insert(
            0, 'sarracenia.flowcb.gather.message.Message')

        if hasattr(self.o, 'post_exchange'):
            self.plugins['load'].insert(
                0, 'sarracenia.flowcb.post.message.Message')



import logging
import requests
import base64

import datetime
import os
import sys
import time

from datetime import date

import sarracenia
from sarracenia.flowcb.scheduled import Scheduled

#
# Support for features inventory mechanism.
#
from sarracenia.featuredetection import features

features['wiski'] = { 'modules_needed': [ 'kiwis_pie' ], 'Needed': True,
        'lament' : 'cannot poll sites that provide data using WISKI API' ,
        'rejoice' : 'can poll sites that provide data with WISKI API' }

try:
    from kiwis_pie import KIWIS
    sarracenia.features['wiski']['present'] = True
except:
    sarracenia.features['wiski']['present'] = False



logger = logging.getLogger(__name__)




class Wiski(Scheduled):
    """
      
       Plugin to Poll a WISKIS server that uses  ( https://www.kisters.net/wiski )
       Uses the kiwis_pie ( https://github.com/amacd31/kiwis_pie ) to do that.

       In the credentials.conf file, need an authentication  entry like:

       https://user:password@theserver.com

       and in the config file:

       pollUrl https://theserver.com

       wiski polling parametrization (these are defined by and passed to kiwis_pie):

       wiski_ts_length -- how long a timeseries to request (default is 24 hours)
       wiski_ts_name   -- name of the timeseries (not used currently.)
       wiski_ts_parameterTypeName -- Air Temperature.

       inherits the following settings from Scheduled (interval overrides the other two.)

       scheduled_hour    -- a list of hours (0-59) to poll.
       scheduled_minute -- a list of minutes (0-59) to poll. 
       scheduled_interval  -- just a time interval between polls.

       2023/09 plugin by Peter Silva, based on original tidbits by Mohamed Rehouma.
    """

    def __init__(self,options):
        super().__init__(options,logger)

        # use self.o.PollUrl for credentials.
        self.details = None
        if self.o.pollUrl is not None:
            ok, self.details = sarracenia.config.Config.credentials.get(
                self.o.pollUrl)

        if self.o.pollUrl is None or self.details == None:
            logger.error("pollUrl option incorrect or missing\n")
            sys.exit(1)

        # assert: now self.details.url.username/password are set.
        self.basic_auth = "Basic " + base64.b64encode(f"{self.details.url.username}:{self.details.url.password}".encode()).decode()

        if self.details.url.port:
            self.main_url = self.details.url.scheme + "://" + self.details.url.hostname + ":" + self.details.url.port + "/KiWIS/KiWIS"
        else:
            self.main_url = self.details.url.scheme + "://" + self.details.url.hostname + "/KiWIS/KiWIS"

        self.host = self.details.url.hostname
        self.token = None
        #self.token = self.submit_tokenization_request()

        # meteorological parameter settings.
        self.o.add_option( 'wiski_ts_length', 'duration', '24h' )
        self.o.add_option( 'wiski_ts_name', 'str', 'caw_Cmd' )
        self.o.add_option( 'wiski_ts_parameterTypeName', 'str', 'Air Temperature' )
        
        self.ts_length = datetime.timedelta( seconds=self.o.wiski_ts_length )
        
        logger.info( f"main_url: {self.main_url} timeseries_length={self.ts_length}, writing to {self.o.directory}" )


    
    def submit_tokenization_request(self):
        # ECCC User Definition
       
        logger.info("requesting a new token")

        body_string = "grant_type=client_credentials&scope=empty"
        
        # Request Token using HTTP POST request
        token_url = "https://" + self.host + "/KiWebPortal/rest/auth/oidcServer/token"
        headers = {
            "Content-Type": "application/x-www-form-urlencoded",
            "Authorization": self.basic_auth,
            "Host": self.host,
            "Cache-Control": "no-cache",
            "Connection": "keep-alive",
        }
        
        response = requests.post(token_url, data=body_string, headers=headers)
        
        str_response_status = response.status_code
        str_response_text = response.text
        
        # Process response
        token_array = str_response_text.split('"')
        if str_response_status == 201:
            submit_tokenization_request = token_array[3]
        else:
            logger.error( f"tokenization_request status: {str_response_text}" )
            submit_tokenization_request = "Error"
        
        return submit_tokenization_request

    def gather(self): # placeholder
        
        messages=[]

        self.wait_until_next()

        while (1):
            if self.stop_requested:
                return messages
        
            self.token = self.submit_tokenization_request()
            authenticated_url = self.main_url
            headers = {
                "Authorization" : f"Bearer {self.token}",
                "Content-Type" : "application/json"
            }
        
            response = requests.get(authenticated_url,headers=headers)
        
            logger.info(response)

            if response.status_code == 200:
                break
            else:
                logger.info( f"request failed. Status code: {response.status_code}: {response.text}" )
        
        k = KIWIS(self.main_url, headers=headers )

        now = datetime.datetime.fromtimestamp(time.time(),datetime.timezone.utc)
        then = now - self.ts_length

        logger.info( f"stations: {k.get_station_list().station_id} " )

        for station_id in k.get_station_list().station_id:

            if self.stop_requested:
                return messages

            timeseries = k.get_timeseries_list(station_id = station_id ).ts_id
            #logger.info( f"looping over the timeseries: {timeseries}" )

            #timeseries = k.get_timeseries_list(station_id = station_id, ts_name =self.o.wiski_ts_name, parametertype_name = self.o.wiski_ts_parameterTypeName ).ts_id
            for ts_id in timeseries:
                # writing files on windows is quite painful, so many illegal characters.
                if sys.platform.startswith( "win" ):
                    fname = f"{self.o.directory}{os.sep}ts_{ts_id}_{station_id}__{str(now).replace(' ','T')}.csv" 
                    fname = fname[0:3]+fname[3:].replace(':','_').replace('.','_',1).replace('+','_').replace('-','_')
                else:
                    fname = f"{self.o.directory}{os.sep}ts_{ts_id}_{station_id}_{str(then).replace(' ','T')}_{str(now).replace(' ','T')}.csv" 
    
                logger.info( f"Timeseries {ts_id} for station_id {station_id} to be written to: {fname}" )
                f=open(fname,'w')
                #ts=k.get_timeseries_values(ts_id = ts_id, to = date(2023,1,31), **{'from': date(2023,1,1)})
                ts=k.get_timeseries_values(ts_id = ts_id, to = now, **{'from': then})
                if len(ts) > 0:
                    ts.to_csv(f)
                else:
                    logger.info( f"no data to write to {f}")
                f.close() 
                messages.append( sarracenia.Message.fromFileData( fname, self.o, os.stat(fname) ) )
    
        return messages

if __name__ == '__main__':

    import sarracenia.config
    import types
    import sarracenia.flow

    options = sarracenia.config.default_config()
    flow = sarracenia.flow.Flow(options)
    flow.o.scheduled_interval= 5
    flow.o.pollUrl = "https://kiwis.opg.com"
    if sys.platform.startswith( "win" ):
        flow.o.directory = "C:\\temp\wiski"
    else:
        flow.o.directory = "/tmp/wiski"
    logging.basicConfig(level=logging.DEBUG)

    me = Wiski(flow.o)
    me.gather()




#
# This file is part of sarracenia.
# The sarracenia suite is Free and is proudly provided by the Government of Canada
# Copyright (C) Her Majesty The Queen in Right of Canada, 2008-2022
#

import sarracenia.moth
import copy


import datetime
import html.parser
import logging
import os
import paramiko

import sarracenia
from sarracenia.featuredetection import features

if features['ftppoll']['present']:
    import dateparser
    import pytz

import sarracenia.config
from sarracenia.flowcb import FlowCB
import sarracenia.transfer
import stat
import sys, time

logger = logging.getLogger(__name__)


def file_size_fix(str_value) -> int:
    try:

        factor = 1
        if str_value[-1] in 'bB': str_value = str_value[:-1]
        elif str_value[-1] in 'kK': factor = 1024
        elif str_value[-1] in 'mM': factor = 1024 * 1024
        elif str_value[-1] in 'gG': factor = 1024 * 1024 * 1024
        elif str_value[-1] in 'tT': factor = 1024 * 1024 * 1024 * 1024
        if str_value[-1].isalpha(): str_value = str_value[:-1]

        fsize = float(str_value) * factor
        isize = int(fsize)

    except:
        logger.debug("bad size %s" % str_value)
        return -1

    return isize


file_type_dict = {
    'l': 0o120000,  # symbolic link
    's': 0o140000,  # socket file
    '-': 0o100000,  # regular file
    'b': 0o060000,  # block device
    'd': 0o040000,  # directory
    'c': 0o020000,  # character device
    'p': 0o010000  # fifo (named pipe)
}


def modstr2num(m) -> int:
    mode = 0
    if (m[0] == 'r'): mode += 4
    if (m[1] == 'w'): mode += 2
    if (m[2] == 'x'): mode += 1
    return mode


def filemode(self, modstr) -> int:
    mode = 0
    mode += file_type_dict[modstr[0]]
    mode += modstr2num(modstr[1:4]) << 6
    mode += modstr2num(modstr[4:7]) << 3
    mode += modstr2num(modstr[7:10])
    return mode


def fileid(self, id) -> int:
    if id.isnumeric():
        return int(id)
    else:
        return None


class Poll(FlowCB):
    """
      The Poll flow callback class implements the main logic for polling remote resources.
      the *poll* routine returns a list of messages for new files to be filtered. 

      when instantiated with options, the options honoured include:

      * pollUrl - the URL of the server to be polled.

      * post_baseURL - parameter for messages to be returned. Also used to look
        up credentials to help subscribers with retrieval.

      * masks - These are the directories at the pollUrl to poll.
        derived from the accept/reject clauses, but filtering should happen later.
        entire directories are listed at this point.

      * timezone - interpret listings from an FTP server as being in the given timezone 
        (as per `pytz <pypi.org/project/pytz>`_

      * chmod - used to identify the minimum permissions to accept for a file to
        be included in a polling result.

      * identity_method - parameter for how to build identity checksum for messages.
        as these are usually remote files, the default is typically "cod" (calculate on download)

      * rename - parameter used to to put in messages built to specify the rename field contents.

      * options are passed to sarracenia.Transfer classes for their use as well.


      Poll uses sarracenia.transfer (ftp, sftp, https, etc... )classes to 
      requests lists of files using those protocols using built-in logic.  

      Internally, Poll normalizes the listings received by placing them into paramiko.SFTPAttributes
      metadata records (similar to stat records) and builds a Sarracenia.Message from them.
      The *poll* routine does one pass of this, returning a list of Sarracenia.Messages.

      To customize: 

      * one can add new sarracenia.transfer protocols, each implementing the *ls* entry point
        to be compatible with this polling routine, ideally the entry point would return a
        list of paramiko.SFTPAttributes for each file in a directory listing.
        This can be used to implement polling of structured remote resources such as S3 or webdav.

      * one can deal with different formats of HTTP pages by overriding the handle_data entry point,
        as done in `nasa_mls_nrt.py <nasa_mls_nrt.py>`_ plugin 

      * for traditional file servers, the listing format should be decypherable with the built-in processing.

      * sftp file servers provide paramiko.SFTPAttributes naturally which are timezone agnostic.

      * for some FTP servers, one may need to specify the *timezone* option to override the UTC default.

      * If there are problems with date or line formats, one can sub-class poll, and override only the on_line
        routine to deal with that.


    """
    def handle_starttag(self, tag, attrs):
        if tag == "table":
            self.tabular_format=True
        elif tag == "tr":
            self.table_column=0
        elif tag == "td":
            self.table_column +=1
        else:
            for attr in attrs:
                c, n = attr
                if c == "href":
                    self.myfname = n.strip().strip('\t')

    def handle_data(self, data):
        """
           routine called from html.parser to deal with a single line.
           if the line is about a file, then create a new entry for it
           with a metadata available from SFTPAttributes.

           example lines:

              from hpfx.collab.science.gc.ca:
                  20230113T00Z_MSC_REPS_HGT_ISBL-0850_RLatLon0.09x0.09_PT000H.grib2   2023-01-13 03:49  5.2M
              from https://data.cosmic.ucar.edu/suominet/nrt/ncConus/y2023/
                  CsuPWVh_2023.011.22.00.0060_nc                     11-Jan-2023 23:58     47K

           this can be overridden by subclassing to deal with new web sites.

           Other web servers put their file indices in a tabular format,  where there is a number
           of cells per row:
           <tr><td></td><td href=filename>filename</td><td>yyyy-mm-dd hh:mm</td><td>size</td>
           This handle_data supports both formats... 
           the tabular format is provided by a vanilla apache2 on a debian derived system.

        """
        logger.debug( f"handling_data {data} column={self.table_column}" )

        if self.tabular_format:
            if self.table_column == 2:
                self.myfname=data
                return
            elif self.table_column != 3:
                return 
            sdate=data.strip()
        else:
            if self.myfname == None: return
            if self.myfname == data: return

            words = data.split()

            if len(words) != 3:
                self.myfname = None
                return

            sdate = words[0] + ' ' + words[1]
 
        if len(sdate) < 10:
            return

        entry = paramiko.SFTPAttributes()

        t=None
        for f in [  '%d-%b-%Y %H:%M', '%Y-%m-%d %H:%M' ]:
            logger.debug( f" try parsing +{sdate}+ using {f}" )
            try:
                t = time.strptime(sdate, f)
                break
            except Exception as Ex:
                pass

        if t:
            mydate = time.strftime('%b %d %H:%M', t)
            entry.st_mtime = time.mktime(t)

        # size is rounded, need a way to be more precise.
        #entry.st_size = file_size_fix(words[-1])

        if self.myfname[-1] != '/':
            entry.st_mode = 0o755
        else:
            entry.st_mode = stat.S_IFDIR | 0o755

        self.entries[self.myfname] = entry
        self.myfname = None

    def on_html_page(self, data) -> dict:
        """
           called once per directory or page of HTML, invokes html.parser, returns
           a dictionary of file entries.
        """
        self.entries = {}
        self.myfname = None
        self.tabular_format=False
        self.table_column=0

        self.parser.feed(data)
        self.parser.close()
        
        return self.entries

    def on_html_parser_init(self):
        # HTML Parsing stuff.
        self.parser = html.parser.HTMLParser()
        self.parser.handle_starttag = self.handle_starttag
        self.parser.handle_data = self.handle_data

    """
      HTML Parsing begine

    """

    def __init__(self, options,class_logger=None):

        super().__init__(options,class_logger)

        # check pollUrl

        self.details = None
        if self.o.pollUrl is not None:
            ok, self.details = sarracenia.config.Config.credentials.get(
                self.o.pollUrl)

        if self.o.pollUrl is None or self.details == None:
            logger.error("pollUrl option incorrect or missing\n")
            sys.exit(1)
        
        if self.o.post_baseUrl is None:
            self.o.post_baseUrl = self.details.url.geturl()
            if self.o.post_baseUrl[-1] != '/': self.o.post_baseUrl += '/'
            if self.o.post_baseUrl.startswith('file:'):
                self.o.post_baseUrl = 'file:'
            if self.details.url.password:
                self.o.post_baseUrl = self.o.post_baseUrl.replace(
                    ':' + self.details.url.password, '')

        self.o.sendTo = self.o.pollUrl

        self.dest = sarracenia.transfer.Transfer.factory(
            self.details.url.scheme, self.o)

        if self.dest is None:
            logger.critical("unsupported polling protocol")

        # rebuild mask as pulls instructions
        # pulls[directory] = [mask1,mask2...]

        #self.pulls = {}
        #for mask in self.o.masks:
        #    pattern, maskDir, maskFileOption, mask_regexp, accepting, mirror, strip, pstrip, flatten = mask
        #    logger.debug(mask)
        #    if not maskDir in self.pulls:
        #        self.pulls[maskDir] = []
        #    self.pulls[maskDir].append(mask)

        self.metricsReset()
        self.on_html_parser_init()

    def metricsReset(self) -> None:
        self.metrics = { 'transferRxBytes': 0 }

    def metricsReport(self) -> dict:
        return self.metrics

    def cd(self, path):
        try:
            self.dest.cd(path)
            return True
        except:
            logger.warning("sr_poll/cd: could not cd to directory %s" % path)
        return False

    def filedate(self, line):

        if not features['ftppoll']['present']:
           logger.error('need dateparser library to deal with polling of ftp servers, no date parsed')
           return 0

        line_split = line.split()
        file_date = line_split[5] + " " + line_split[6] + " " + line_split[7]
        current_date = datetime.datetime.now(pytz.utc)
        # case 1: the date contains '-' implies the date is in 1 string not 3 seperate ones, and H:M is also provided
        if "-" in file_date: file_date = line_split[5] + " " + line_split[6]
        standard_date_format = dateparser.parse(
            file_date,
            settings={
                'RELATIVE_BASE': datetime.datetime(current_date.year, 1, 1),
                'TIMEZONE': self.o.timezone,  #turn this into an option - should be EST for mtl
                'TO_TIMEZONE': 'UTC'
            })
        if standard_date_format is not None:
            # case 2: the year was not given, it is defaulted to 1900. Must find which year (this one or last one).
            if standard_date_format.month - current_date.month >= 6:
                standard_date_format = standard_date_format.replace(
                    year=(current_date.year - 1))
        timestamp = datetime.datetime.timestamp(standard_date_format)
        return timestamp

    def on_line(self, line) -> paramiko.SFTPAttributes:
        """
           default line processing, converts a file listing into an SFTPAttributes.
           does nothing if input is already an SFTPAttributes item, returning it unchanged.
           verifies that file is accessible (based on self.o.permDefault pattern to establish minimum permissions.)
        """
        if type(line) is paramiko.SFTPAttributes:
            sftp_obj = line
        elif type(line) is str and len(line.split()) < 7:
            # assume windows...
            parts = line.split()
            sftp_obj = paramiko.SFTPAttributes()
            ldate = dateparser.parse( ' '.join(parts[0:2]), settings={ 'TIMEZONE': self.o.timezone, 'TO_TIMEZONE':'UTC' } )
            sftp_obj.st_mtime = ldate.timestamp()
            sftp_obj.st_size = file_size_fix(parts[2])
            sftp_obj.longname = ' '.join(line[3:])
            sftp_obj.st_mode = 0o644 # just make it work... no permission info provided.
            #logger.info( f"windows line parsing result: {sftp_obj}")
        elif type(line) is str and len(line.split()) > 7:

            parts = line.split()
            sftp_obj = paramiko.SFTPAttributes()
            sftp_obj.st_mode = filemode(self,parts[0])
            sftp_obj.st_uid = fileid(self,parts[2])
            sftp_obj.st_gid = fileid(self,parts[3])

            if file_size_fix(parts[4]) >= 0: # normal linux/unix ftp server case.
                sftp_obj.st_size = file_size_fix(parts[4])
                sftp_obj.filename = line[8:]
                sftp_obj.st_mtime = self.filedate(line)
            else: # university of wisconsin (some special file system? has third ownship field before size) 
                sftp_obj.st_size = file_size_fix(parts[5])
                sftp_obj.filename = line[9:]
                sftp_obj.st_mtime = self.filedate(line[1:])

            sftp_obj.longname = sftp_obj.filename


        # assert at this point we have an sftp_obj...
        if 'sftp_obj' in locals() and ((sftp_obj.st_mode
                                        & self.o.permDefault) == self.o.permDefault):
            return sftp_obj
        else:
            return None

    def lsdir(self):

        try:
            ls = self.dest.ls()

            if type(ls) is bytes:
                self.metrics["transferRxBytes"] += len(ls)
                ls = self.on_html_page(ls.decode('utf-8'))

            new_ls = {}
            new_dir = {}
            # del ls['']  # For some reason with FTP the first line of the ls causes an index out of bounds error becuase it contains only "total ..." in line_mode.py

            # apply selection on the list

            for f in ls:
                logger.debug( f"line to parse: {f}" )
                matched = False
                line = ls[f]

                line = self.on_line(line)
                if (line is None) or (line == ""):
                    continue
                if stat.S_ISDIR(line.st_mode):
                    new_dir[f] = line
                else:
                    new_ls[f] = line

            return True, new_ls, new_dir
        except Exception as e:
            logger.warning("dest.lsdir: Could not ls directory")
            logger.debug("Exception details:", exc_info=True)

        return False, {}, {}

    def poll_directory(self, pdir):
        #logger.debug("poll_directory %s %s" % (pdir))
        npost = 0
        msgs = []

        # cd to that directory

        logger.debug(" cd %s" % pdir)
        ok = self.cd(pdir)
        if not ok: return []

        # ls that directory

        ok, file_dict, dir_dict = self.lsdir()
        if not ok: return []

        filelst = file_dict.keys()
        desclst = file_dict

        logger.debug("poll_directory: new files found %d" % len(filelst))

        # post poll list

        msgs.extend(self.poll_list_post(pdir, dir_dict, dir_dict.keys()))

        msgs.extend(self.poll_list_post(pdir, desclst, filelst))

        # poll in children directory

        sdir = sorted(dir_dict.keys())
        for d in sdir:
            if d == '.' or d == '..': continue

            #d_lspath = lspath + '_' + d
            d_pdir = pdir + os.sep + d

            msgs.extend(self.poll_directory(d_pdir))

        return msgs

    def poll_file_post(self, desc, destDir, remote_file):

        FileOption = None
        for mask in self.pulllst:
            pattern, maskDir, maskFileOption, mask_regexp, accepting, mirror, strip, pstrip, flatten = mask
            if mask_regexp.match(remote_file) and accepting:
                FileOption = maskFileOption

        path = destDir + '/' + remote_file

        # posting a localfile
        if self.o.post_baseUrl.startswith('file:'):
            if os.path.isfile(path) or os.path.islink(path):
                try:
                    lstat = sarracenia.stat(path)
                except:
                    lstat = None

                ok = sarracenia.Message.fromFileInfo(path, self.o, lstat)
                if os.path.islink(path):
                    if 'size' in msg:
                        del msg['size']
                    if not self.o.follow_symlinks:
                        try: 
                            ok['fileOp'] = { 'link': os.readlink(path) } 
                            if 'Identity' in msg:
                                 del ok['Identity']
                        except:
                            logger.error("cannot read link %s message dropped" % path)
                            logger.debug('Exception details: ', exc_info=True)
                            ok=None
                return ok

        post_relPath = destDir + '/' + remote_file

        logger.debug('desc: type: %s, value: %s' % (type(desc), desc))

        if type(desc) == str:
            line = desc.split()
            st = paramiko.SFTPAttributes()
            st.st_size = file_size_fix(line[4])
            # actionally only need to convert normalized time to number here...
            # just being lazy...
            lstime = dateparser.parse(line[5] + " " + line[6]).timestamp()
            st.st_mtime = lstime
            st.st_atime = lstime

            desc = st

        msg = sarracenia.Message.fromFileInfo(post_relPath, self.o, desc)

        if stat.S_ISDIR(desc.st_mode):
            if 'mkdir' not in self.o.fileEvents:
                return None

            msg['fileOp'] = { 'directory':'' }
             
        elif stat.S_ISLNK(desc.st_mode):
            if 'link' not in self.o.fileEvents:
                return None

            if not self.o.follow_symlinks:
                try: 
                    msg['fileOp'] = { 'link': self.dest.readlink(path) }
                except:
                    logger.error("cannot read link %s message dropped" % post_relPath)
                    logger.debug('Exception details: ', exc_info=True)
                    return None

        if 'create' not in self.o.fileEvents and 'modify' not in self.o.fileEvents:
            return None

        if self.o.identity_method and (',' in self.o.identity_method):
            m, v = self.o.identity_method.split(',')
            msg['identity'] = {'method': m, 'value': v}

        # If there is a file operation, and it isn't a rename, then some fields are irrelevant/wrong.
        if 'fileOp' in msg and 'rename' not in msg['fileOp']: 
            if 'Identity' in msg:
                del msg['Indentity']
            if 'size' in msg:
                del msg['size']

        this_rename = self.o.rename

        # FIX ME generalized fileOption
        if FileOption is not None:
            parts = FileOption.split('=')
            option = parts[0].strip()
            if option == 'rename' and len(parts) == 2:
                this_rename = parts[1].strip()

        if this_rename is not None and this_rename[-1] == '/':
            this_rename += remote_file

        if this_rename is not None:
            msg['rename'] = this_rename

        return [msg]

    def poll_list_post(self, destDir, desclst, filelst):

        n = 0
        msgs = []

        for idx, remote_file in enumerate(filelst):
            desc = desclst[remote_file]
 
            new_msgs = self.poll_file_post(desc, destDir, remote_file)
            if new_msgs:
                msgs.extend(new_msgs)
        return msgs

    # =============
    # for all directories, get urls to post
    # if True is returned it means : no sleep, retry on return
    # False means, go to sleep and retry after sleep seconds
    # =============

    def poll(self) -> list:

        # General Attributes

        self.pulllst = []

        msgs = []
        # number of post files

        npost = 0

        # connection did not work

        try:
            self.dest.connect()
        except:
            logger.error("sr_poll/post_new_url: unable to connect to %s" %
                         self.o.pollUrl)
            logger.debug('Exception details: ', exc_info=True)
            nap=15
            logger.error("Sleeping {nap} secs and retry")
            time.sleep(nap)
            return []

        # loop on all directories where there are pulls to do

        for destDir in self.o.path:

            # setup of poll directory info

            #self.pulllst = self.pulls[destDir]

            #path = destDir
            #path = path.replace('${', '')
            #path = path.replace('}', '')
            #path = path.replace('/', '_')
            #lsPath = self.o.cfg_run_dir + os.sep + 'ls' + path

            currentDir = self.o.variableExpansion(destDir)

            if currentDir == '': currentDir = destDir
            msgs.extend(self.poll_directory(currentDir))
            logger.debug('poll_directory returned: %s' % len(msgs))

        # close connection

        try:
            self.dest.close()
        except:
            pass

        return msgs



#!/usr/bin/python3

"""
Plugin to download from NASA using an Earthdata account
========================================================

This plugin enables authentication for URLs posted by the ``poll_NASA_CMR.py`` plugin, and other NASA data
sources that are compatible with bearer tokens that come from https://urs.earthdata.nasa.gov.

This plugin is required because it's not possible to use a username and password to authenticate with the URL
to be downloaded. Either you need to use a bearer token, or login to https://urs.earthdata.nasa.gov and download
your file using the authenticated session, which has cookies to track your login. 

Sarracenia supports bearer token authentication, so the default downloading code can be used. But, NASA bearer
tokens expire after 90 days. This plugin handles dynamically creating and retrieving the bearer tokens from the
https://urs.earthdata.nasa.gov API.

For every message processed by this plugin, it will add a credential with a valid bearer token matching the message's
baseUrl to Sarracenia's in-memory credential database.

  
Configurable Options:
----------------------

``earthdataUrl``:
^^^^^^^^^^^^^^^^^^^^

    Default is ``https://urs.earthdata.nasa.gov``. If you want to override it, you can specify the
    ``earthdataUrl https://your-endpoint.example.com`` option.

How to set up your download config:
--------------------------------
 
    Add ``callback accept.auth_NASA_Earthdata``, in your subscribe, sarra or other download config.  

    Add ``https://username:password@urs.earthdata.nasa.gov/`` to your ``credentials.conf`` file.  
    
    Optional: set ``acceptSizeWrong True`` in the sarra/subscribe config to suppress the WARNING message
     about a file being downloaded with no length given.

    For examples, see https://github.com/MetPX/sarracenia/tree/main/sarracenia/examples/subscribe files
    named ``*nasa_earthdata*.conf``. 

Change log:
-----------

    - 2023-10-10: first attempt at this plugin. The old v2 code re-implemented downloading, using a session with
      stored cookies. This should be better, because it uses the native sr3 download code.
"""

import sarracenia
import datetime
import logging
import requests

logger = logging.getLogger(__name__)

class Auth_nasa_earthdata(sarracenia.flowcb.FlowCB):
    def __init__(self, options):
        super().__init__(options, logger)
        
        # Allow setting a logLevel *only* for this plugin in the config file:
        # set accept.auth_NASA_Earthdata.logLevel debug
        if hasattr(self.o, 'logLevel'):
            logger.setLevel(self.o.logLevel.upper())

        logger.debug("plugin: Download_NASA_Earthdata __init__")

        self.o.add_option('earthdataUrl', kind='str', default_value='https://urs.earthdata.nasa.gov')

        if self.o.earthdataUrl[-1] == '/':
            self.o.earthdataUrl = self.o.earthdataUrl.strip('/')

        self._token = None
        self._token_expires = None # format MM/DD/YYYY, e.g. 12/25/2023

        # end __init__

    def after_accept(self, worklist):
        """ Doesn't actually do any downloading. First, get or create a bearer token from
            https://urs.earthdata.nasa.gov by logging in with the username and password from credentials.conf.
            Then adds the bearer token to Sarracenia's credentials DB for the message's baseUrl. This will allow
            the file to be downloaded from msg['baseUrl']+msg['relPath'] using the bearer token.
        """

        # It's not clear what time the token expires on the expiry date. If today = expiry date, then try to get a
        # token every time this runs. If it's not expired yet, we'll get the same token from the API and can try 
        # to check again on the next run. If it is expired, we should get a brand new token.
        today = datetime.datetime.utcnow().strftime("%m/%d/%Y")
        if self._token_expires and today == self._token_expires:
            logger.info("the token ending with ...{self._token[-5:]} expires today ({self._token_expires})")
            self._token = None
            self._token_expires = None

        # Get a token from the NASA API, if there isn't one already
        if not self._token:
            # Try to get a new token
            if not self.get_earthdata_token():
                logger.error(f"Failed to retrieve Bearer token from {self.o.earthdataUrl}. " + 
                             f"Can't download {msg['baseUrl']}{msg['relPath']}")

        for msg in worklist.incoming:
            
            # If the credential already exists and the bearer_token matches, don't need to do anything
            ok, details = self.o.credentials.get(msg['baseUrl'])
            token_already_in_creds = False
            try: 
                token_already_in_creds = (ok and details.bearer_token == self._token)
                if token_already_in_creds:
                    logger.debug(f"Token for {msg['baseUrl']} already in credentials database")
            except:
                token_already_in_creds = False

            if not token_already_in_creds:
                logger.info(f"Token for {msg['baseUrl']} not in credentials database. Adding it!")
                # Add the new bearer token to the internal credentials db. If the credential is already in the db, it will
                # be replaced which is desirable.
                cred = sarracenia.credentials.Credential(urlstr=msg['baseUrl'])
                cred.bearer_token = self._token
                self.o.credentials.add(msg['baseUrl'], details=cred)


    def create_earthdata_token(self, auth: requests.auth.HTTPBasicAuth) -> bool:
        """ Create a new Earthdata token.
        """
        self._token = None
        self._token_expires = None

        try:
            # Create a new token
            resp = requests.post(self.o.earthdataUrl + "/api/users/token", auth=auth)
            if resp.status_code != 200:
                logger.error(f"Failed to create a new token! Code: {resp.status_code} Info: {resp.text}")
                return False
            
            # If we got 200, success!
            resp_j = resp.json()
            # logger.debug(f"Here's the response: {resp_j}")

            self._token = resp_j['access_token']
            self._token_expiry = resp_j['expiration_date']
            logger.info(f"Successfully created new token! " + 
                        f"Token ends with ...{self._token[-5:]} and expires on {self._token_expiry}")
            return True
            
        except Exception as e:
            logger.error(f"Failed to create a new token! {e}")
            logger.debug("details:", exc_info=True)
            return False

    def get_earthdata_token(self) -> bool:
        """ Try to retrieve a token from the Earthdata account. If there is no token, it will create a new one.
        https://urs.earthdata.nasa.gov/documentation/for_users/user_token
        """
        self._token = None
        self._token_expires = None

        # From config file
        endpoint = self.o.earthdataUrl

        # Credentials - from Earthdata endpoint URL
        ok, details = self.o.credentials.get(endpoint)
        if not ok:
            logger.error(f"credential lookup failed for {endpoint}. Check your credentials.conf file!")
            return False
        creds = details.url
        username = creds.username
        password = creds.password

        auth = requests.auth.HTTPBasicAuth(username, password)

        try:
            # Try to get an existing token
            resp = requests.get(self.o.earthdataUrl + "/api/users/tokens", auth=auth)
            if resp.status_code != 200:
                logger.error(f"Failed to login to NASA Earthdata. Code: {resp.status_code} Info: {resp.text}")
                return False
            
            # If we got 200, we either have an empty response (user has 0 tokens), or we have a token
            resp_j = resp.json()
            # logger.debug(f"Here's the response: {resp_j}, {len(resp_j)}")

            if len(resp_j) >= 1: 
                self._token = resp_j[0]['access_token']
                self._token_expiry = resp_j[0]['expiration_date']
                logger.info(f"There is/are {len(resp_j)} token(s) in {username}'s Earthdata account. " + 
                            f"Using the token that ends with ...{self._token[-5:]} and expires on {self._token_expiry}")
                return True
            # Valid response, but no tokens in the account. Need to create one.
            else:
                logger.info(f"There are {len(resp_j)} tokens in {username}'s Earthdata account. A new token will be created.")
                return self.create_earthdata_token(auth)
            
        except Exception as e:
            logger.error(f"Failed to login to NASA Earthdata. {e}")
            logger.debug("details:", exc_info=True)
            return False

        


"""
Plugin postoverride.py:
    Override message header for products that are posted. This can be useful or necessary 
    when re-distributing beyond the original intended destinations.

Example:
    for example company A delivers to their own DMZ server. ACME is a client of them,
    and so subscribes to the ADMZ server, but the to_cluster=ADMZ, when ACME downloads, they
    need to override the destination to specify the distribution within ACME.
    * postOverride to_clusters ACME
    * postOverrideDel from_cluster

Usage: 
    flowcb sarracenia.flowcb.accept.postoverride.PostOverride
    postOverride x y
    postOverrideDel z

"""

import logging
import copy
from sarracenia.flowcb import FlowCB

logger = logging.getLogger(__name__)


class PostOverride(FlowCB):
    def __init__(self, options):
        super().__init__(options,logger)
        self.o.add_option('postOverride', 'list')
        self.o.add_option('postOverrideDel', 'list')

        if self.o.postOverride != None:
            logger.info('postOverride settings: %s' % self.o.postOverride)
        if self.o.postOverrideDel != None:
            logger.info('postOverrideDel settings: %s' % self.o.postOverrideDel)

    def after_accept(self, worklist):
        for message in worklist.incoming:
            
            if self.o.postOverride != None:
                for o in self.o.postOverride:
                    (osetting, ovalue) = o.split()
                    logger.debug('postOverride applying key:%s value:%s' % (osetting, ovalue))
                    message[osetting] = ovalue

            if self.o.postOverrideDel != None:
                for od in self.o.postOverrideDel:
                    if od in message:
                        logger.debug('postOverride deleting key:%s ' % od)
                        del message[od]



from sarracenia.flow import Flow
import logging

logger = logging.getLogger(__name__)

default_options = {'acceptUnmatched': True, 'nodupe_ttl': 0}


class Shovel(Flow):
    """
       * subscribe to some messages.
       * post them somewhere else.
    """
    def __init__(self, options):

        super().__init__(options)
        self.plugins['load'].insert(
            0, 'sarracenia.flowcb.gather.message.Message')
        self.plugins['load'].insert(0,
                                    'sarracenia.flowcb.post.message.Message')
